(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory RPO
imports Term_Order_Impl Lexicographic_Extension Term_Impl Reduction_Pair SN_Order_Carrier Multiset_Extension
begin


(* precedence, max arity, and two terms are given, result is pair (strict, non-strict) *)
function
  lpo :: "('f \<Rightarrow> nat \<Rightarrow> 'f \<Rightarrow> nat \<Rightarrow> bool \<times> bool) \<times> ('f \<Rightarrow> bool) \<Rightarrow> nat \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> bool \<times> bool"
where
  "lpo _ _ (Var x) (Var y) = (False, x = y)" |
  "lpo pr _ (Var x) (Fun g ts) = (False, ts = [] \<and> (snd pr) g)" |
  "lpo pr n (Fun f ss) (Var y) = (let con = (\<exists> s \<in> set ss. snd (lpo pr n s (Var y))) in (con,con))" |
  "lpo pr n (Fun f ss) (Fun g ts) = (
    if (\<exists> s \<in> set ss. snd (lpo pr n s (Fun g ts)))
       then (True,True)
       else (let (prs,prns) = (fst pr) f (length ss) g (length ts) in 
         if prns \<and> (\<forall> t \<in> set ts. fst (lpo pr n (Fun f ss) t))
          then if prs then (True,True) 
            else lex_ext (lpo pr n) n ss ts
          else (False,False)))"
by pat_completeness auto


termination lpo 
proof (relation "measures [\<lambda> (pr,n,s,t). size s + size t]")
  fix "pr" n and f :: 'f and ss :: "('f,'v)term list" and g :: 'f and ts :: "('f,'v)term list" and x xa xb xc i
  assume "i < length ss" and "i < length ts"
  then obtain s t where si: "s = ss ! i" and s: "s \<in> set ss" and ti: "t = ts ! i" and t: "t \<in> set ts" by auto
  from s have s1: "size s \<le> list_size size ss" by (induct ss, auto)
  from t have t1: "size t \<le> list_size size ts" by (induct ts, auto)
  from s1 t1 si ti show "((pr, n, ss ! i, ts ! i), pr, n, Fun f ss, Fun g ts) \<in> measures [\<lambda> (pr,n,s,t). size s + size t]"
    by auto
qed (auto simp: size_simps)



(* precedence, max arity, and two terms are given, result is pair (strict, non-strict) *)
function
  mpo :: "('f \<Rightarrow> nat \<Rightarrow> 'f \<Rightarrow> nat \<Rightarrow> bool \<times> bool) \<times> ('f \<Rightarrow> bool) \<Rightarrow> nat \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> bool \<times> bool"
where
  "mpo _ _ (Var x) (Var y) = (False, x = y)" |
  "mpo pr _ (Var x) (Fun g ts) = (False, ts = [] \<and> (snd pr) g)" |
  "mpo pr n (Fun f ss) (Var y) = (let con = (\<exists> s \<in> set ss. snd (mpo pr n s (Var y))) in (con,con))" |
  "mpo pr n (Fun f ss) (Fun g ts) = (
    if (\<exists> s \<in> set ss. snd (mpo pr n s (Fun g ts)))
       then (True,True)
       else (let (prs,prns) = (fst pr) f (length ss) g (length ts) in 
         if prns \<and> (\<forall> t \<in> set ts. fst (mpo pr n (Fun f ss) t))
          then if prs then (True,True) 
            else mul_ext (mpo pr n) ss ts
          else (False,False)))"
by pat_completeness auto


termination mpo 
proof (relation "measures [\<lambda> (pr,n,s,t). size s + size t]")
  fix "pr" n and f :: 'f and ss :: "('f,'v)term list" and g :: 'f and ts :: "('f,'v)term list" and x xa y xb x'
  assume s: "xb \<in> set ss" and t: "x' \<in> set ts"
  from s have s1: "size xb \<le> list_size size ss" by (induct ss, auto)
  from t have t1: "size x' \<le> list_size size ts" by (induct ts, auto)
  from s1 t1 s t show "((pr, n, xb, x'), pr, n, Fun f ss, Fun g ts) \<in> measures [\<lambda> (pr,n,s,t). size s + size t]"
    by auto
qed (auto simp: size_simps)

lemma lpo_stri_imp_nstri[rule_format]:  "fst (lpo pr n s t) \<longrightarrow>
  snd (lpo pr n s t)"
proof (induct rule: lpo.induct[of "\<lambda> pr n s t. fst (lpo pr n s t) \<longrightarrow> snd (lpo pr n s t)"])
  case (4 "pr" n f ss g ts)
  obtain prc prl where prec: "pr = (prc,prl)"  by (cases "pr", auto)
  obtain  s ns where  prc: "prc f (length ss) g (length ts) = (s,ns)" by (cases "prc f (length ss) g (length ts)", auto)
  show ?case 
    by (cases "length ss = length ts \<or> length ts \<le> n", auto simp: Let_def lex_ext_stri_imp_nstri prec prc) 
qed (auto simp: Let_def)

lemma mpo_stri_imp_nstri[rule_format]:  "fst (mpo pr n s t) \<longrightarrow>
  snd (mpo pr n s t)"
proof (induct rule: mpo.induct[of "\<lambda> pr n s t. fst (mpo pr n s t) \<longrightarrow> snd (mpo pr n s t)"])
  case (4 "pr" n f ss g ts)
  obtain prc prl where prec: "pr = (prc,prl)"  by (cases "pr", auto)
  obtain  s ns where  prc: "prc f (length ss) g (length ts) = (s,ns)" by (cases "prc f (length ss) g (length ts)", auto)
  {
    assume "fst (mul_ext (mpo (prc, prl) n) ss ts)"
    hence "snd (mul_ext (mpo (prc, prl) n) ss ts)"
      unfolding mul_ext_def
      by (auto simp: Let_def)
  }
  thus ?case 
    by (simp add: Let_def prec prc)
qed (auto simp: Let_def)

locale rpo_pr =
  fixes prc :: "'f \<Rightarrow> nat \<Rightarrow> 'f \<Rightarrow> nat \<Rightarrow> bool \<times> bool"
  and   prl :: "'f \<Rightarrow> bool"
  assumes prc_refl: "prc f n f n = (False,True)"
  and     prc_SN: "SN {((f,n),(g,m)) | f n g m. fst (prc f n g m)}"
  and     prl: "prl g \<Longrightarrow> snd (prc f n g 0) = True"
  and     prl3: "prl f \<Longrightarrow> snd (prc f 0 g 0) \<Longrightarrow> prl g"
  and  prc_compat: "prc f n g m = (s1,ns1) \<Longrightarrow> prc g m h k = (s2,ns2) \<Longrightarrow> prc f n h k = (s,ns) \<Longrightarrow> (ns1 \<and> ns2 \<longrightarrow> ns) \<and> (ns1 \<and> s2 \<longrightarrow> s) \<and> (s1 \<and> ns2 \<longrightarrow> s)"
begin

lemma prl2: assumes g: "prl g" shows "fst (prc g 0 f n) = False"
proof(rule ccontr)
  assume "\<not> ?thesis"
  then obtain b where gf: "prc g 0 f n = (True,b)" by (cases "prc g 0 f n", auto)
  obtain b1 b2 where gg: "prc g 0 g 0 = (b1,b2)" by (cases "prc g 0 g 0", auto)
  obtain b' where fg: "prc f n g 0 = (b',True)" using prl[OF g, of f n] by (cases "prc f n g 0", auto)
  from prc_compat[OF gf fg gg] gg have gg: "fst (prc g 0 g 0)" by auto
  with SN_imp_irreflexive[OF prc_SN] show False by blast
qed

abbreviation "pr" where "pr \<equiv> (prc,prl)"
abbreviation lpo_pr where "lpo_pr \<equiv> lpo pr"
abbreviation mpo_pr where "mpo_pr \<equiv> mpo pr"

lemma lpo_nstri_refl: 
  shows "snd (lpo_pr n s s)"
proof (induct s)
  case (Fun f ss)
  have rec1: "snd (lex_ext (lpo_pr n) n ss ss)"
    by (rule all_nstri_imp_lex_nstri, auto simp: Fun)
  have rec2: "\<forall> s \<in> set ss. fst (lpo_pr n (Fun f ss) s)"
  proof
    fix s
    assume s: "s \<in> set ss"
    with Fun have rec: "snd (lpo_pr n s s)" by auto
    with s show "fst (lpo_pr n (Fun f ss) s)"
    proof (cases s)
      case (Fun g ts)
      from s rec have "\<exists> t \<in> set ss. snd (lpo_pr n t s)" by auto
      with Fun show ?thesis by auto
    next
      case (Var x)
      from s rec have "\<exists> t \<in> set ss. snd (lpo_pr n t s)" by auto
      with Var show ?thesis by auto
    qed
  qed
  from rec1 rec2 show ?case by (auto simp: Let_def prc_refl)
qed simp

lemma mpo_nstri_refl: 
  shows "snd (mpo_pr n s s)"
proof (induct s)
  case (Fun f ss)
  have rec1: "snd (mul_ext (mpo_pr n) ss ss)"
    unfolding mul_ext_def Let_def snd_conv
  proof (rule ns_mul_ext_refl)
    show "locally_reflexive {(x,y). snd (mpo_pr n x y)} (multiset_of ss)"
      unfolding locally_reflexive_def
    proof (intro allI impI)
      fix s
      assume "s \<in># multiset_of ss"
      hence "s \<in> set ss" unfolding mem_set_multiset_eq .
      from Fun[OF this] show "(s,s) \<in> {(x,y). snd (mpo_pr n x y)}" by simp
    qed
  qed
  have rec2: "\<forall> s \<in> set ss. fst (mpo_pr n (Fun f ss) s)"
  proof
    fix s
    assume s: "s \<in> set ss"
    with Fun have rec: "snd (mpo_pr n s s)" by auto
    with s show "fst (mpo_pr n (Fun f ss) s)"
    proof (cases s)
      case (Fun g ts)
      from s rec have "\<exists> t \<in> set ss. snd (mpo_pr n t s)" by auto
      with Fun show ?thesis by auto
    next
      case (Var x)
      from s rec have "\<exists> t \<in> set ss. snd (mpo_pr n t s)" by auto
      with Var show ?thesis by auto
    qed
  qed
  from rec1 rec2 show ?case by (auto simp: Let_def prc_refl)
qed simp

lemma supt_imp_lpo_stri: assumes "s \<rhd> t" 
  shows "fst (lpo_pr n s t)"
  using assms
proof (induct)
  case (arg s ss f)
  have refl: "snd (lpo_pr n s s)" by (rule lpo_nstri_refl)
  from arg refl have fact: "\<exists> t \<in> set ss. snd (lpo_pr n t s)" by auto
  thus ?case 
    by (cases s, auto)
next
  case (subt s ss t f)
  from subt lpo_stri_imp_nstri[of "pr" n s t] have "\<exists> s \<in> set ss. snd (lpo_pr n s t)" by auto
  thus ?case
    by (cases t, auto)
qed

lemma supt_imp_mpo_stri: assumes "s \<rhd> t" 
  shows "fst (mpo_pr n s t)"
  using assms
proof (induct)
  case (arg s ss f)
  have refl: "snd (mpo_pr n s s)" by (rule mpo_nstri_refl)
  from arg refl have fact: "\<exists> t \<in> set ss. snd (mpo_pr n t s)" by auto
  thus ?case 
    by (cases s, auto)
next
  case (subt s ss t f)
  from subt mpo_stri_imp_nstri[of "pr" n s t] have "\<exists> s \<in> set ss. snd (mpo_pr n s t)" by auto
  thus ?case
    by (cases t, auto)
qed


lemma lpo_nstri_mono: assumes rel: "snd (lpo_pr n s t)"
  shows "snd (lpo_pr n (Fun f (bef @ s # aft)) (Fun f (bef @ t # aft)))"
proof -
  let ?ss = "bef @ s # aft"
  let ?ts = "bef @ t # aft"
  let ?s = "Fun f ?ss"
  let ?len = "Suc (length bef + length aft)"
  have aft_bef: "\<forall> b \<in> set bef \<union> set aft. fst (lpo_pr n ?s b)"
  proof (intro ballI)
    fix b
    assume elem: "b \<in> set bef \<union> set aft"
    show "fst (lpo_pr n ?s b)"
    proof (rule supt_imp_lpo_stri, cases "b \<in> set bef")
      assume "b \<in> set bef"
      thus "?s \<rhd> b" by (auto simp: supt.intros)
    next
      assume bbef: "b \<notin> set bef"
      show "?s \<rhd> b" 
      proof (cases "b \<in> set aft")
        assume "b \<notin> set aft"
        thus "?s \<rhd> b" using bbef elem by blast
      qed (auto simp: supt.intros)
    qed
  qed
  from rel have "\<exists> s \<in> set ?ss. snd (lpo_pr n s t)" by (auto)
  hence t: "fst (lpo_pr n ?s t)" by (cases t, auto)
  let ?map = "map (\<lambda> (s,t). lpo_pr n s t)"
  have lex_ext: "snd (lex_ext (lpo_pr n) n ?ss ?ts)"
  proof (rule all_nstri_imp_lex_nstri[of ?ts _ ?ss, simplified], intro allI impI)
    fix i
    assume i: "i < ?len"
    show "snd (lpo_pr n (?ss ! i) (?ts ! i))" (is "?nstri i")
    proof (cases "i = length bef")
      case True
      with rel show "?nstri i" by auto
    next
      case False
      show ?thesis 
      proof (cases "i < length bef")
        case True
        with lpo_nstri_refl[of n] show "?nstri i" by (simp add: nth_append)
      next
        case False
        with `i \<noteq> length bef` have "i - length bef > 0" by arith 
        then obtain j where i: "i - length bef = Suc j" by (cases "i - length bef", auto)
        with False lpo_nstri_refl[of n "aft ! j"] show "?nstri i" by (simp add: nth_append)
      qed
    qed
  qed
  show ?thesis by (simp only: lpo.simps Let_def, cases "\<exists> s \<in> set ?ss. snd (lpo_pr n s (Fun f ?ts))", simp, simp add: prc_refl t aft_bef lex_ext[simplified])
qed


lemma mpo_nstri_mono: assumes rel: "snd (mpo_pr n s t)"
  shows "snd (mpo_pr n (Fun f (bef @ s # aft)) (Fun f (bef @ t # aft)))"
proof -
  let ?ss = "bef @ s # aft"
  let ?ts = "bef @ t # aft"
  let ?s = "Fun f ?ss"
  let ?len = "Suc (length bef + length aft)"
  have aft_bef: "\<forall> b \<in> set bef \<union> set aft. fst (mpo_pr n ?s b)"
  proof (intro ballI)
    fix b
    assume elem: "b \<in> set bef \<union> set aft"
    show "fst (mpo_pr n ?s b)"
    proof (rule supt_imp_mpo_stri, cases "b \<in> set bef")
      assume "b \<in> set bef"
      thus "?s \<rhd> b" by (auto simp: supt.intros)
    next
      assume bbef: "b \<notin> set bef"
      show "?s \<rhd> b" 
      proof (cases "b \<in> set aft")
        assume "b \<notin> set aft"
        thus "?s \<rhd> b" using bbef elem by blast
      qed (auto simp: supt.intros)
    qed
  qed
  from rel have "\<exists> s \<in> set ?ss. snd (mpo_pr n s t)" by (auto)
  hence t: "fst (mpo_pr n ?s t)" by (cases t, auto)
  let ?map = "map (\<lambda> (s,t). mpo_pr n s t)"
  from length_append have sslts: "length ?ss = length ?ts" by simp
  have "(multiset_of ?ss, multiset_of ?ts) \<in> ns_mul_ext {(x, y). snd (mpo_pr n x y)}
       {(x, y). fst (mpo_pr n x y)}"
  proof (rule all_ns_ns_mul_ext)
  from sslts show "length ?ss = length ?ts" by assumption
  {
    fix i assume i_bd: "i < length ?ts"
    have "(?ss ! i, ?ts  ! i) \<in> {(x, y). snd (mpo_pr n x y)}"
    proof (cases "i < length bef")
    case True
      with append_Cons_nth_left[OF this] and mpo_nstri_refl
       show ?thesis by simp
    next
    case False note Hineq = this
      thus ?thesis proof (cases "i - length bef")
      case 0
        with False have "i = length bef" by simp
        with nth_append_length and rel show ?thesis by simp
      next
      case (Suc j)
        from nth_append and False
         have "\<And>u. (bef @ u # aft) ! i = (u # aft) ! (i - length bef)" by metis
        with Suc and mpo_nstri_refl show ?thesis by simp
      qed
    qed
  }
  thus "\<forall>i. i < length ?ts \<longrightarrow> (?ss ! i, ?ts ! i) \<in> {(x, y). snd (mpo_pr n x y)}" by simp
  qed
  hence mul_ext: "snd (mul_ext (mpo_pr n) ?ss ?ts)" unfolding mul_ext_def by simp
  show ?thesis by (simp only: mpo.simps Let_def, cases "\<exists> s \<in> set ?ss. snd (mpo_pr n s (Fun f ?ts))", simp, simp add: prc_refl t aft_bef mul_ext[simplified])
qed

lemma lpo_stri_mono: 
  assumes rel: "fst (lpo_pr n s t)"
  shows "fst (lpo_pr n (Fun f (bef @ s # aft)) (Fun f (bef @ t # aft)))"
proof -
  let ?ss = "bef @ s # aft"
  let ?ts = "bef @ t # aft"
  let ?s = "Fun f ?ss"
  have aft_bef: "\<forall> b \<in> set bef \<union> set aft. fst (lpo_pr n ?s b)"
  proof (intro ballI)
    fix b
    assume elem: "b \<in> set bef \<union> set aft"
    show "fst (lpo_pr n ?s b)"
      proof (rule supt_imp_lpo_stri, cases "b \<in> set bef")
        assume "b \<in> set bef" 
        thus "?s \<rhd> b" by (auto simp: supt.intros)
      next 
        assume bbef: "b \<notin> set bef"
        show "?s \<rhd> b" 
        proof (cases "b \<in> set aft")
	  assume "b \<notin> set aft"
	  thus "?s \<rhd> b" using bbef elem by blast
        qed (auto simp: supt.intros)
      qed
  qed
  from lpo_stri_imp_nstri[OF rel] have "\<exists> s \<in> set ?ss. snd (lpo_pr n s t)" by auto
  hence t: "fst (lpo_pr n ?s t)" by (cases t, auto)
  let ?map = "map (\<lambda> (s,t). lpo_pr n s t)"
  obtain stri nstri where st: "lpo_pr n s t = (stri,nstri)" by force
  with rel have stri: "stri = True" by simp
  {
    fix i
    assume i: "i < length bef"
    hence "snd (lpo_pr n (?ss ! i) (?ts ! i))" by (simp add: lpo_nstri_refl[of n] nth_append)
  }
  hence lex_ext: "fst (lex_ext (lpo_pr n) n ?ss ?ts)"
    by (simp add: lex_ext_iff, intro exI[of _ "length bef"], simp add: st stri)
  show ?thesis by (simp only: lpo.simps Let_def, cases "\<exists> s \<in> set ?ss. snd (lpo_pr n s (Fun f ?ts))", simp, simp add: prc_refl t aft_bef lex_ext)
qed



lemma mpo_stri_mono: 
  assumes rel: "fst (mpo_pr n s t)"
  shows "fst (mpo_pr n (Fun f (bef @ s # aft)) (Fun f (bef @ t # aft)))"
proof -
  let ?ss = "bef @ s # aft"
  let ?ts = "bef @ t # aft"
  let ?s = "Fun f ?ss"
  have aft_bef: "\<forall> b \<in> set bef \<union> set aft. fst (mpo_pr n ?s b)"
  proof (intro ballI)
    fix b
    assume elem: "b \<in> set bef \<union> set aft"
    show "fst (mpo_pr n ?s b)"
      proof (rule supt_imp_mpo_stri, cases "b \<in> set bef")
        assume "b \<in> set bef" 
        thus "?s \<rhd> b" by (auto simp: supt.intros)
      next 
        assume bbef: "b \<notin> set bef"
        show "?s \<rhd> b" 
        proof (cases "b \<in> set aft")
	  assume "b \<notin> set aft"
	  thus "?s \<rhd> b" using bbef elem by blast
        qed (auto simp: supt.intros)
      qed
  qed
  from mpo_stri_imp_nstri[OF rel] have "\<exists> s \<in> set ?ss. snd (mpo_pr n s t)" by auto
  hence t: "fst (mpo_pr n ?s t)" by (cases t, auto)
  let ?map = "map (\<lambda> (s,t). mpo_pr n s t)"
  obtain stri nstri where st: "mpo_pr n s t = (stri,nstri)" by force
  with rel have stri: "stri = True" by simp
  let ?ba = "bef @ aft"
  have Hrew: "\<And>t. remove_nth (length bef) (bef @ t # aft) = ?ba" by simp
  have bef_bd: "\<And>t. length bef < length (bef @ t # aft)" by simp
  with mpo_nstri_refl
   have "\<forall>i. i < length ?ba \<longrightarrow> (?ba ! i, ?ba ! i) \<in> {(x, y). snd (mpo_pr n x y)}"
   by auto
  from all_ns_ns_mul_ext[OF _ this]
   have H: "(multiset_of ?ba, multiset_of ?ba) \<in> ns_mul_ext  {(x, y). snd (mpo_pr n x y)} {(x, y). fst (mpo_pr n x y)}"
   by metis
  from rel have s_s_t: "\<forall>b. b \<in># {#t#} \<longrightarrow> (\<exists>a. a \<in># {#s#} \<and> (a,b) \<in> {(x,y). fst (mpo_pr n x y)})" by simp
  from remove_nth_soundness[OF bef_bd[of s]] and Hrew and nth_append_length[of bef s aft]
   and insert_DiffM2[OF nth_mem_multiset_of[OF bef_bd[of s]]]
   have H1: "multiset_of ?ba + {#s#} = multiset_of ?ss" by auto
  from remove_nth_soundness[OF bef_bd[of t]] and Hrew and nth_append_length[of bef t aft]
   and insert_DiffM2[OF nth_mem_multiset_of[OF bef_bd[of t]]]
   have H2: "multiset_of ?ba + {#t#} = multiset_of ?ts" by auto
  have "{#s#} \<noteq> {#}" by simp
  from ns_s_mul_ext_union_multiset_l[OF H this s_s_t] and H1 and H2
   have "(multiset_of ?ss, multiset_of ?ts)
     \<in> s_mul_ext  {(x, y). snd (mpo_pr n x y)} {(x, y). fst (mpo_pr n x y)}" by simp
  hence mul_ext: "fst (mul_ext (mpo_pr n) ?ss ?ts)" unfolding mul_ext_def by simp
  show ?thesis by (simp only: mpo.simps Let_def, cases "\<exists> s \<in> set ?ss. snd (mpo_pr n s (Fun f ?ts))", simp, simp add: prc_refl t aft_bef mul_ext)
qed


lemma lpo_stable: 
  "(fst (lpo_pr n s t) \<longrightarrow> fst (lpo_pr n (s \<cdot> \<sigma>) (t \<cdot> \<sigma>))) \<and> (snd (lpo_pr n s t) \<longrightarrow> snd (lpo_pr n (s \<cdot> \<sigma>) (t \<cdot> \<sigma>)))"
  (is "?p s t")
proof -
  have "pr = pr \<longrightarrow> ?p s t"
  proof (induct rule: lpo.induct[of "\<lambda> p n s t. p = pr \<longrightarrow> (fst (lpo_pr n s t) \<longrightarrow> fst (lpo_pr n (s \<cdot> \<sigma>) (t \<cdot> \<sigma>))) \<and> (snd (lpo_pr n s t) \<longrightarrow> snd (lpo_pr n (s \<cdot> \<sigma>) (t \<cdot> \<sigma>)))"])
    case (1 _ n x y)
    show ?case by (auto simp: lpo_nstri_refl)
  next
    case (2 _ n x g ts)
    show ?case
    proof -
      show ?thesis 
      proof (cases "\<sigma> x")
        case (Var y)
        show ?thesis by (auto simp: Var)
      next
        case (Fun f ss)
        show ?thesis 
        proof (cases "prl g")
          case False thus ?thesis  by auto
        next
          case True
          obtain s ns where "prc f (length ss) g 0 = (s,ns)" by (cases "prc f (length ss) g 0", auto)        
          with prl[OF True, of f "length ss"] have prc: "prc f (length ss) g 0 = (s, True)" by auto
          show ?thesis by (auto simp: Fun prc all_nstri_imp_lex_nstri[of "[]", simplified]) 
        qed
      qed
    qed
  next
    case (3 p n f ss y)
    {
      assume "\<exists> s \<in> set ss. snd (lpo_pr n s (Var y))" and p: "p = pr"
      from this obtain s where s: "s \<in> set ss" and "snd (lpo_pr n s (Var y))" by force
      with 3 p have "snd (lpo_pr n (s \<cdot> \<sigma>) (Var y \<cdot> \<sigma>))" by auto
      with s have "\<exists> s' \<in> set (map (\<lambda> s'. s' \<cdot> \<sigma>) ss). snd (lpo_pr n s' (Var y \<cdot> \<sigma>))" by auto
    }
    thus ?case by (cases "Var y \<cdot> \<sigma>", auto simp: Let_def)
  next
    case (4 p n f ss g ts)
    let ?map = "map (\<lambda> s'. s' \<cdot> \<sigma>)"
    let ?t = "Fun g ts"
    let ?s = "Fun f ss"
    let ?ssig = "Fun f ss \<cdot> \<sigma>"
    let ?tsig = "Fun g ts \<cdot> \<sigma>"
    let ?ss = "?map ss"
    let ?ts = "?map ts"
    show ?case
    proof (cases "p = pr")
      case True note p = this
      show ?thesis
      proof (cases "\<exists> s \<in> set ss. snd (lpo_pr n s ?t)")
        case True
        from this obtain s where s: "s \<in> set ss" and "snd (lpo_pr n s ?t)" by auto
        with 4 p have "snd (lpo_pr n (s \<cdot> \<sigma>) ?tsig)" by auto
        with s have "\<exists> s' \<in> set ?ss. snd (lpo_pr n s' ?tsig)" by auto
        thus ?thesis by auto
      next
        case False
        hence not1: "\<not> (\<exists> s \<in> set ss. snd (lpo_pr n s ?t))" by force
        show ?thesis 
        proof (cases "snd (lpo_pr n ?s ?t)")
          case False
          with lpo_stri_imp_nstri have "\<not> (fst (lpo_pr n ?s ?t))" by blast
          with False show ?thesis by auto
        next
          case True
          with not1 have smallTs: "\<forall> t \<in> set ts. fst (lpo_pr n ?s t)" by (cases ?thesis, auto)
          obtain s ns where prc: "prc f (length ss) g (length ts) = (s,ns)" by (cases "prc f (length ss) g (length ts)", auto)
          from True smallTs have ge: "ns = True" by (cases ?thesis, auto simp: Let_def not1 prc simp del: list_all_ex)
          note prc = prc[unfolded ge]
          have "\<forall> t \<in> set ts. fst (lpo_pr n ?ssig (t \<cdot> \<sigma>))"
            by (intro ballI, rule 4(2)[THEN mp[OF _ p], THEN conjunct1, THEN mp[OF _ smallTs[THEN bspec]], unfolded p, OF not1], auto simp: prc)
          hence smallTssig: "\<forall> t \<in> set ?ts. fst (lpo_pr n ?ssig t)" by auto
          show ?thesis 
          proof (cases s)
	    case True
	    with smallTssig not1 show ?thesis by (auto simp: prc)
          next
	    case False
            hence "s = False" by simp
            note prc = prc[unfolded this]
	    with not1 smallTs True have snd_lex: "snd (lex_ext (lpo_pr n) n ss ts)" by auto
	    let ?ind = "\<lambda> i. (fst (lpo_pr n (ss ! i) (ts ! i)) \<longrightarrow> fst (lpo_pr n (ss ! i \<cdot> \<sigma>) (ts ! i \<cdot> \<sigma>))) \<and> (snd (lpo_pr n (ss ! i) (ts ! i)) \<longrightarrow> snd (lpo_pr n (ss ! i \<cdot> \<sigma>) (ts ! i \<cdot> \<sigma>)))"
            have ind: "\<And> i. \<lbrakk>i < length ss; i < length ts\<rbrakk> \<Longrightarrow> ?ind i" 
              by (rule 4(3)[unfolded p, OF not1 refl, simplified], auto simp: prc smallTs)
            from snd_lex have eq_small: "length ss = length ts \<or> length ts \<le> n" by (simp add: lex_ext_iff)
            from ind snd_lex have snd_lex2: "snd (lex_ext (lpo_pr n) n ?ss ?ts)"
              by (simp add: lex_ext_iff eq_small, auto)
	    have nstri: "snd (lpo_pr n ?ssig ?tsig)" by (simp add: Let_def prc snd_lex smallTssig[simplified] snd_lex2)
	    {
	      assume "fst (lpo_pr n ?s ?t)"
	      with not1 smallTs prc snd_lex have "fst (lex_ext (lpo_pr n) n ss ts)" by auto
              with ind have fst_lex2: "fst (lex_ext (lpo_pr n) n ?ss ?ts)"
                by (simp add: lex_ext_iff eq_small, auto)
	      have "fst (lpo_pr n ?ssig ?tsig)" by (simp add: Let_def prc snd_lex smallTssig[simplified] fst_lex2)
	    }
	    with nstri show ?thesis by simp
          qed
        qed
      qed
    next
      case False
      thus ?thesis by simp
    qed
  qed
  thus ?thesis by simp
qed


lemma mpo_stable: fixes \<sigma> :: "('f,'v)subst"
  shows "(fst (mpo_pr n s t) \<longrightarrow> fst (mpo_pr n (s \<cdot> \<sigma>) (t \<cdot> \<sigma>))) \<and> (snd (mpo_pr n s t) \<longrightarrow> snd (mpo_pr n (s \<cdot> \<sigma>) (t \<cdot> \<sigma>)))"
  (is "?p s t")
proof -
  have "pr = pr \<longrightarrow> ?p s t"
  proof (induct rule: mpo.induct[of "\<lambda> p n s t. p = pr \<longrightarrow> (fst (mpo_pr n s t) \<longrightarrow> fst (mpo_pr n (s \<cdot> \<sigma>) (t \<cdot> \<sigma>))) \<and> (snd (mpo_pr n s t) \<longrightarrow> snd (mpo_pr n (s \<cdot> \<sigma>) (t \<cdot> \<sigma>)))"])
    case (1 _ n x y)
    show ?case by (auto simp: mpo_nstri_refl)
  next
    case (2 _ n x g ts)
    show ?case
    proof -
      show ?thesis 
      proof (cases "\<sigma> x")
        case (Var y)
        show ?thesis by (auto simp: Var)
      next
        case (Fun f ss)
        show ?thesis 
        proof (cases "prl g")
          case False thus ?thesis  by auto
        next
          case True
          obtain s ns where "prc f (length ss) g 0 = (s,ns)" by (cases "prc f (length ss) g 0", auto)        
          with prl[OF True, of f "length ss"] have prc: "prc f (length ss) g 0 = (s, True)" by auto
          from ns_mul_ext_bottom have "snd (mul_ext (mpo_pr n) ss [])" by (simp add: mul_ext_def)
          thus ?thesis by (auto simp: Fun prc all_nstri_imp_mul_nstri[of "[]", simplified])
        qed
      qed
    qed
  next
    case (3 p n f ss y)
    {
      assume "\<exists> s \<in> set ss. snd (mpo_pr n s (Var y))" and p: "p = pr"
      from this obtain s where s: "s \<in> set ss" and "snd (mpo_pr n s (Var y))" by force
      with 3 p have "snd (mpo_pr n (s \<cdot> \<sigma>) (Var y \<cdot> \<sigma>))" by auto
      with s have "\<exists> s' \<in> set (map (\<lambda> s'. s' \<cdot> \<sigma>) ss). snd (mpo_pr n s' (Var y \<cdot> \<sigma>))" by auto
    }
    thus ?case by (cases "Var y \<cdot> \<sigma>", auto simp: Let_def)
  next
    case (4 p n f ss g ts)
    let ?map = "map (\<lambda> s'. s' \<cdot> \<sigma>)"
    let ?t = "Fun g ts"
    let ?s = "Fun f ss"
    let ?ssig = "Fun f ss \<cdot> \<sigma>"
    let ?tsig = "Fun g ts \<cdot> \<sigma>"
    let ?ss = "?map ss"
    let ?ts = "?map ts"
    show ?case
    proof (cases "p = pr")
      case True note p = this
      show ?thesis
      proof (cases "\<exists> s \<in> set ss. snd (mpo_pr n s ?t)")
        case True
        from this obtain s where s: "s \<in> set ss" and "snd (mpo_pr n s ?t)" by auto
        with 4 p have "snd (mpo_pr n (s \<cdot> \<sigma>) ?tsig)" by auto
        with s have "\<exists> s' \<in> set ?ss. snd (mpo_pr n s' ?tsig)" by auto
        thus ?thesis by auto
      next
        case False
        hence not1: "\<not> (\<exists> s \<in> set ss. snd (mpo_pr n s ?t))" by force
        show ?thesis 
        proof (cases "snd (mpo_pr n ?s ?t)")
          case False
          with mpo_stri_imp_nstri have "\<not> (fst (mpo_pr n ?s ?t))" by blast
          with False show ?thesis by auto
        next
          case True
          with not1 have smallTs: "\<forall> t \<in> set ts. fst (mpo_pr n ?s t)" by (cases ?thesis, auto)
          obtain s ns where prc: "prc f (length ss) g (length ts) = (s,ns)" by (cases "prc f (length ss) g (length ts)", auto)
          from True smallTs have ge: "ns = True" by (cases ?thesis, auto simp: Let_def not1 prc simp del: list_all_ex)
          note prc = prc[unfolded ge]
          have "\<forall> t \<in> set ts. fst (mpo_pr n ?ssig (t \<cdot> \<sigma>))"
            by (intro ballI, rule 4(2)[THEN mp[OF _ p], THEN conjunct1, THEN mp[OF _ smallTs[THEN bspec]], unfolded p, OF not1], auto simp: prc)
          hence smallTssig: "\<forall> t \<in> set ?ts. fst (mpo_pr n ?ssig t)" by auto
          show ?thesis 
          proof (cases s)
	    case True
	    with smallTssig not1 show ?thesis by (auto simp: prc)
          next
	    case False
            hence "s = False" by simp
            note prc = prc[unfolded this]
	    with not1 smallTs True have snd_mul: "snd (mul_ext (mpo_pr n) ss ts)" by auto
	    let ?ind = "\<lambda> s t. (fst (mpo_pr n s t) \<longrightarrow> fst (mpo_pr n (s \<cdot> \<sigma>) (t \<cdot> \<sigma>))) \<and> (snd (mpo_pr n s t) \<longrightarrow> snd (mpo_pr n (s \<cdot> \<sigma>) (t \<cdot> \<sigma>)))"
            have ind: "\<And> s t. s \<in> set ss \<Longrightarrow> t \<in> set ts \<Longrightarrow> ?ind s t"
              by (rule 4(3)[unfolded p, OF not1 refl, simplified], auto simp: prc smallTs)
            have snd_mul2: "snd (mul_ext (mpo_pr n) ?ss ?ts)"
              by (rule nstri_mul_ext_map, insert snd_mul ind, auto)
	    have nstri: "snd (mpo_pr n ?ssig ?tsig)" by (simp add: Let_def prc smallTssig[simplified] snd_mul2)
	    {
	      assume "fst (mpo_pr n ?s ?t)"
	      with not1 smallTs prc snd_mul have fst_mul: "fst (mul_ext (mpo_pr n) ss ts)" by auto
              have fst_mul2: "fst (mul_ext (mpo_pr n) ?ss ?ts)"
                by (rule stri_mul_ext_map, insert fst_mul ind, auto)
	      have "fst (mpo_pr n ?ssig ?tsig)" by (simp add: Let_def prc smallTssig[simplified] fst_mul2)
	    }
	    with nstri show ?thesis by simp
          qed
        qed
      qed
    next
      case False
      thus ?thesis by simp
    qed
  qed
  thus ?thesis by simp
qed

lemma lpo_least_1: assumes "prl f" 
  shows "snd (lpo_pr n t (Fun f []))"
proof (cases t, simp add: assms)
  case (Fun g ts)
  obtain s ns where "prc g (length ts) f 0 = (s,ns)" by (cases "prc g (length ts) f 0", auto)
  with prl[OF assms, of g "length ts"] have prc: "prc g (length ts) f 0 = (s,True)" by auto
  show ?thesis unfolding Fun by (simp add: lex_ext_least_1 prc )
qed

lemma mpo_least_1: assumes "prl f"
  shows "snd (mpo_pr n t (Fun f []))"
proof (cases t)
case (Var y) thus ?thesis by (simp add: assms)
next
case (Fun g ts)
  obtain s ns where "prc g (length ts) f 0 = (s,ns)" by (cases "prc g (length ts) f 0", auto)
  with prl[OF assms, of g "length ts"] have prc: "prc g (length ts) f 0 = (s,True)" by auto
  from ns_mul_ext_bottom have "snd (mul_ext (mpo_pr n) ts [])" unfolding mul_ext_def by simp
 thus ?thesis using assms and Fun and prc by simp
qed

lemma lpo_least_2: assumes "prl f" 
  shows "\<not> fst (lpo_pr n (Fun f []) t)"
proof (cases t, simp)
  case (Fun g ts)
  obtain s ns where "prc f 0 g (length ts) = (s,ns)" by (cases "prc f 0 g (length ts)", auto)
  with prl2[OF assms, of g "length ts"] have prc: "prc f 0 g (length ts) = (False,ns)" by auto
  show ?thesis unfolding Fun by (simp add: Let_def lex_ext_least_2 prc)
qed

lemma mpo_least_2: assumes "prl f" 
  shows "\<not> fst (mpo_pr n (Fun f []) t)"
proof (cases t, simp)
  case (Fun g ts)
  obtain s ns where "prc f 0 g (length ts) = (s,ns)" by (cases "prc f 0 g (length ts)", auto)
  with prl2[OF assms, of g "length ts"] have prc: "prc f 0 g (length ts) = (False,ns)" by auto
  from s_mul_ext_bottom_strict have "\<not> fst (mul_ext (mpo_pr n) [] ts)"
  unfolding mul_ext_def by simp
  thus ?thesis unfolding Fun by (simp add: Let_def lex_ext_least_2 prc)
qed

lemma lpo_least_3: assumes "prl f" 
  and "snd (lpo_pr n (Fun f []) t)"
  shows "snd (lpo_pr n u t)"
using assms
proof (cases t, simp)
  case (Fun g ts)
  obtain s ns where "prc f 0 g (length ts) = (s,ns)" by (cases "prc f 0 g (length ts)", auto)
  with prl2[OF `prl f`, of g "length ts"] have prc: "prc f 0 g (length ts) = (False,ns)" by auto
  show ?thesis
  proof (cases ts)
    case Nil
    with Fun assms prc have "prc f 0 g (length ts) = (False,True)" by (cases ?thesis, simp, simp add: Let_def)
    with prl3[OF `prl f`, of g] Nil have "prl g" by auto
    show ?thesis unfolding Fun Nil by (rule lpo_least_1[OF `prl g`])
  next
    case (Cons s tts)
    have "\<not> fst (lpo_pr n (Fun f []) s)" by (rule lpo_least_2[OF `prl f`])
    with `snd (lpo_pr n (Fun f []) t)` Fun Cons have False by (simp add: Let_def)
    thus ?thesis ..
  qed
qed

lemma mpo_least_3: assumes "prl f" 
  and "snd (mpo_pr n (Fun f []) t)"
  shows "snd (mpo_pr n u t)"
using assms
proof (cases t, simp)
  case (Fun g ts)
  obtain s ns where "prc f 0 g (length ts) = (s,ns)" by (cases "prc f 0 g (length ts)", auto)
  with prl2[OF `prl f`, of g "length ts"] have prc: "prc f 0 g (length ts) = (False,ns)" by auto
  show ?thesis
  proof (cases ts)
    case Nil
    with Fun assms prc have "prc f 0 g (length ts) = (False,True)" by (cases ?thesis, simp, simp add: Let_def)
    with prl3[OF `prl f`, of g] Nil have "prl g" by auto
    show ?thesis unfolding Fun Nil by (rule mpo_least_1[OF `prl g`])
  next
    case (Cons s tts)
    have "\<not> fst (mpo_pr n (Fun f []) s)" by (rule mpo_least_2[OF `prl f`])
    with `snd (mpo_pr n (Fun f []) t)` Fun Cons have False by (simp add: Let_def)
    thus ?thesis ..
  qed
qed

lemma lpo_compat[rule_format]: 
  fixes s t u :: "('f,'v)term"
  shows "(snd (lpo_pr n s t) \<and> fst (lpo_pr n t u) \<longrightarrow> fst (lpo_pr n s u)) \<and> (fst (lpo_pr n s t) \<and> snd (lpo_pr n t u) \<longrightarrow> fst (lpo_pr n s u)) \<and> (snd (lpo_pr n s t) \<and> snd (lpo_pr n t u) \<longrightarrow> snd (lpo_pr n s u))" (is "?tran s t u")
proof (rule
    wf_induct[OF wf_measures[of "[\<lambda> (s,t,u). size s, \<lambda> (s,t,u). size t, \<lambda> (s,t,u). size u]"], 
    of "\<lambda> (s,t,u). ?tran s t u" "(s,t,u)", simplified], clarify)
  fix s t u :: "('f,'v)term"
  assume ind: "\<forall> s' t' u'. (size s' < size s \<longrightarrow> ?tran s' t' u') \<and> (size s' = size s \<and> (size t' < size t \<or> size t' = size t \<and> size u' < size u) \<longrightarrow> ?tran s' t' u')"
  let ?GR = "\<lambda> s t. fst (lpo_pr n s t)"
  let ?GE = "\<lambda> s t. snd (lpo_pr n s t)"
  show "?tran s t u"
  proof (cases t)
    case (Var x)
    { 
      assume "?GR t u"
      with Var have False by (cases u, auto)
    } 
    moreover
    {
      assume gr: "?GR s t" and ge: "?GE t u"
      have "?GR s u"
      proof (cases u)
	case (Var y)
	with ge `t = Var x` have "t = u" by auto
	with gr show ?thesis by auto
      next
	case (Fun h us)
	with ge Var have us: "us = []" and pri: "prl h" by auto
	from gr Var obtain f ss where s: "s = Fun f ss" by (cases s, auto)
	with gr Var obtain s' where s': "s' \<in> set ss" by (auto simp: Let_def)
	have "?GE s' u" by (simp only: Fun us, rule lpo_least_1, rule pri)
	with s' have "\<exists> s' \<in> set ss. ?GE s' u" by (auto)
	with s Fun show ?thesis by simp
      qed
    }
    moreover
    {
      assume ge1: "?GE s t" and ge2: "?GE t u"
      have "?GE s u"
      proof (cases u)
	case (Var y)
	with ge2 `t = Var x` have "t = u" by auto
	with ge1 show ?thesis by auto
      next
	case (Fun h us)
	with ge2 Var have us: "us = []" and pri: "prl h" by auto
	show ?thesis by (simp only: Fun us, rule lpo_least_1, rule pri) 
      qed
    }
    ultimately show ?thesis by blast
  next
    case (Fun g ts)
    let ?t = "Fun g ts"
    from Fun have t: "t = ?t" .
    show ?thesis 
    proof (cases s)
      case (Var x)
      {
	assume gr: "?GR s t"
	with Var Fun have False by auto
      }
      moreover
      {
	assume ge: "?GE s t" and gr: "?GR t u"
	with Var Fun have pri: "prl g" and "ts = []" by auto
	with gr Fun have False using lpo_least_2[of g n u] by auto
      }
      moreover
      {
	assume ge1: "?GE s t" and ge2: "?GE t u"
	with Var Fun have pri: "prl g" and empty: "ts = []" by auto
	have "?GE s u" by (rule lpo_least_3[of g n u], rule pri, simp add: empty[symmetric] Fun[symmetric], rule ge2)
      }
      ultimately show ?thesis by blast
    next
      case (Fun f ss)
      let ?s = "Fun f ss"
      from Fun have s: "s = ?s" .
      let ?s1 = "\<exists> s' \<in> set ss. ?GE s' t"
      let ?t1 = "\<exists> t' \<in> set ts. ?GE t' u"
      let ?ls = "length ss"
      let ?lt = "length ts"
      obtain ps pns where prc: "prc f ?ls g ?lt = (ps,pns)" by (cases "prc f ?ls g ?lt")
      let ?tran2 = "\<lambda> a b c.  
         (snd (lpo_pr n a b) \<and> fst (lpo_pr n b c) \<longrightarrow> fst (lpo_pr n a c)) \<and>
         (fst (lpo_pr n a b) \<and> snd (lpo_pr n b c) \<longrightarrow> fst (lpo_pr n a c)) \<and> 
         (snd (lpo_pr n a b) \<and> snd (lpo_pr n b c) \<longrightarrow> snd (lpo_pr n a c)) \<and> 
         (fst (lpo_pr n a b) \<and> fst (lpo_pr n b c) \<longrightarrow> fst (lpo_pr n a c))"
      from s have "\<forall> s' \<in> set ss. size s' < size s" by (auto simp: size_simps)
      with ind have ind2: "\<And> s' t' u'. \<lbrakk>s' \<in> set ss\<rbrakk> \<Longrightarrow> ?tran s' t' u'" by blast
      with lpo_stri_imp_nstri have ind3: "\<And> us s' t' u'. \<lbrakk>s' \<in> set ss; t' \<in> set ts; u' \<in> set us\<rbrakk> \<Longrightarrow> ?tran2 s' t' u'" by blast
      {
	assume ge: "?GE s t" and gr: "?GR t u"
	have "?GR s u"
	proof (cases ?s1)
	  case True
	  from this obtain s' where s': "s' \<in> set ss" and ges: "?GE s' t" by auto
	  with s s' ind2[of s' t u, simplified] ges gr have "?GR s' u" by auto
	  hence "?GE s' u" by (rule lpo_stri_imp_nstri)
	  with s' s show ?thesis by (cases u, auto simp: Let_def)
	next
	  case False
	  show ?thesis
	  proof (cases ?t1)
	    case True
	    from this obtain t' where t': "t' \<in> set ts" and ges: "?GE t' u" by auto
	    from t' t False ge s have gr2: "?GR s t'" by (cases "\<forall> t' \<in> set ts. ?GR ?s t'", auto)
	    from t' s t gr2 ges spec[OF spec[OF spec[OF ind]], of s _ u, simplified] show "?GR s u" using size_simps supt.intros by auto
	  next
	    case False
	    from t this gr obtain h us where u: "u = Fun h us" by (cases u, auto simp: Let_def)
	    let ?u = "Fun h us"	    
	    from s t u ge gr have ge: "?GE ?s ?t" and gr: "?GR ?t ?u" by auto
	    from `\<not> ?s1` t ge have st': "\<forall> t' \<in> set ts. ?GR ?s t'" by (cases ?thesis, auto)
	    from `\<not> ?t1` u gr have tu': "\<forall> u' \<in> set us. ?GR ?t u'" by (cases ?thesis, auto)
	    let ?lu = "length us"
            obtain ps2 pns2 where prc2: "prc g ?lt h ?lu = (ps2,pns2)" by (cases "prc g ?lt h ?lu")
            obtain ps3 pns3 where prc3: "prc f ?ls h ?lu = (ps3,pns3)" by (cases "prc f ?ls h ?lu")
	    from `\<not> ?s1` t ge st' have fg: "pns" by (cases ?thesis, auto simp: Let_def prc) 
	    from `\<not> ?t1` u gr tu' have gh: "pns2" by (cases ?thesis, auto simp: Let_def prc2)
            note compat = prc_compat[OF prc prc2 prc3]
            from fg gh compat have fh: "pns3" by simp 
	    from tu' s t ge spec[OF spec[OF ind], of s t, simplified] u have su': "\<forall> u' \<in> set us. ?GR ?s u'" using size_simps supt.intros by auto
	    show ?thesis
	    proof (cases ps3)
	      case True
	      with su' s u fh prc3 show ?thesis by (auto)
	    next
	      case False              
	      from False fg gh compat have nfg: "\<not> ps" and ngh: "\<not> ps2" by blast+
	      from fg `\<not> ?s1` t ge st' prc nfg have lex1: "snd (lex_ext (lpo_pr n) n ss ts)" by auto
	      from gh `\<not> ?t1` u gr tu' prc2 ngh have lex2: "fst (lex_ext (lpo_pr n) n ts us)" by auto
	      from lex1 lex2 lex_ext_compat[OF ind3, of ss ts us] have "fst (lex_ext (lpo_pr n) n ss us)" by auto
	      with s u fh su' prc3 show ?thesis by auto
	    qed
	  qed
	qed
      }
      moreover
      {
	assume gr: "?GR s t" and ge: "?GE t u"
	have "?GR s u"
	proof (cases ?s1)
	  case True
	  from this obtain s' where s': "s' \<in> set ss" and ges: "?GE s' t" by auto
	  with s s' ind2[of s' t u, simplified] ges ge have "?GE s' u" by auto
	  with s' s show ?thesis by (cases u, auto simp: Let_def)
	next
	  case False
	  show ?thesis
	  proof (cases ?t1)
	    case True
	    from this obtain t' where t': "t' \<in> set ts" and ges: "?GE t' u" by auto
	    have tt': "?GE t t'" by (rule lpo_stri_imp_nstri[OF supt_imp_lpo_stri], auto simp: t t' supt.intros)
	    from t' t False gr s have gr2: "?GR s t'" by (cases "\<forall> t' \<in> set ts. ?GR ?s t'", auto)
	    from t' s t gr2 ges spec[OF spec[OF spec[OF ind]], of s _ u, simplified] show "?GR s u" using size_simps by auto
	  next
	    case False
	    from t this ge obtain h us where u: "u = Fun h us" by (cases u, auto simp: Let_def)
	    let ?u = "Fun h us"	    
	    from s t u ge gr have gr: "?GR ?s ?t" and ge: "?GE ?t ?u" by auto
	    from `\<not> ?s1` t gr have st': "\<forall> t' \<in> set ts. ?GR ?s t'" by (cases ?thesis, auto)
	    from `\<not> ?t1` u ge have tu': "\<forall> u' \<in> set us. ?GR ?t u'" by (cases ?thesis, auto)
	    let ?lu = "length us"
            obtain ps2 pns2 where prc2: "prc g ?lt h ?lu = (ps2,pns2)" by (cases "prc g ?lt h ?lu")
            obtain ps3 pns3 where prc3: "prc f ?ls h ?lu = (ps3,pns3)" by (cases "prc f ?ls h ?lu")
	    from `\<not> ?s1` t gr st' have fg: "pns" by (cases ?thesis, auto simp: Let_def prc) 
	    from `\<not> ?t1` u ge tu' have gh: "pns2" by (cases ?thesis, auto simp: Let_def prc2)
            note compat = prc_compat[OF prc prc2 prc3]
            from fg gh compat have fh: "pns3" by simp 
	    from tu' s t lpo_stri_imp_nstri[OF gr] spec[OF spec[OF ind], of s t, simplified] u have su': "\<forall> u' \<in> set us. ?GR ?s u'"
              using size_simps by auto
	    show ?thesis
	    proof (cases ps3)
	      case True
	      with su' s u fh prc3 show ?thesis by (auto)
	    next
	      case False
	      from False fg gh compat have nfg: "\<not> ps" and ngh: "\<not> ps2" by blast+
	      from fg `\<not> ?s1` t gr st' prc nfg have lex1: "fst (lex_ext (lpo_pr n) n ss ts)" by auto
	      from gh `\<not> ?t1` u ge tu' prc2 ngh have lex2: "snd (lex_ext (lpo_pr n) n ts us)" by auto
	      from lex1 lex2 lex_ext_compat[OF ind3, of ss ts us]
	      have "fst (lex_ext (lpo_pr n) n ss us)" by auto
	      with s u fh su' fh prc3 show ?thesis by (auto)
	    qed
	  qed
	qed
      }
      moreover
      {
	assume ge1: "?GE s t" and ge2: "?GE t u" and ngt1: "\<not> ?GR s t" and ngt2: "\<not> ?GR t u"
	have "\<not> ?s1"
	proof
	  assume ?s1
	  with s have "?GR s t" by (cases t, auto)
	  with ngt1 show False ..
	qed
	have "\<not> ?t1"
	proof
	  assume ?t1
	  with t have "?GR t u" by (cases u, auto)
	  with ngt2 show False ..
	qed	
	from t `\<not> ?t1` ge2 obtain h us where u: "u = Fun h us" by (cases u, auto simp: Let_def)
	let ?u = "Fun h us"	    
	from s t u ge1 ge2 have ge1: "?GE ?s ?t" and ge2: "?GE ?t ?u" by auto
	from `\<not> ?s1` t ge1 have st': "\<forall> t' \<in> set ts. ?GR ?s t'" by (cases ?thesis, auto)
	from `\<not> ?t1` u ge2 have tu': "\<forall> u' \<in> set us. ?GR ?t u'" by (cases ?thesis, auto)
	let ?lu = "length us"        
        obtain ps2 pns2 where prc2: "prc g ?lt h ?lu = (ps2,pns2)" by (cases "prc g ?lt h ?lu")
        obtain ps3 pns3 where prc3: "prc f ?ls h ?lu = (ps3,pns3)" by (cases "prc f ?ls h ?lu")
	from `\<not> ?s1` t ge1 st' have fg: "pns" by (cases ?thesis, auto simp: Let_def prc) 
	from `\<not> ?t1` u ge2 tu' have gh: "pns2" by (cases ?thesis, auto simp: Let_def prc2)
        note compat = prc_compat[OF prc prc2 prc3]
        from compat fg gh have fh: pns3 by blast
	from tu' s t ge1 spec[OF spec[OF ind], of s t, simplified] u have su': "\<forall> u' \<in> set us. ?GR ?s u'" using size_simps by auto
	from `\<not> ?s1` st' ge1 ngt1 s t have nfg: "\<not> ps" by (simp add: Let_def, cases ?thesis, simp, cases ps, auto simp: prc fg) 
	from `\<not> ?t1` tu' ge2 ngt2 t u have ngh: "\<not> ps2" by (simp add: Let_def, cases ?thesis, simp, cases ps2, auto simp: prc2 gh)
	from `\<not> ?s1` st' ge1 s t fg nfg have lex1: "snd (lex_ext (lpo_pr n) n ss ts)" by (auto simp: prc)
	from `\<not> ?t1` tu' ge2 t u gh ngh have lex2: "snd (lex_ext (lpo_pr n) n ts us)" by (auto simp: prc2)
	from lex1 lex2 lex_ext_compat[OF ind3, of ss ts us]
	have "snd (lex_ext (lpo_pr n) n ss us)" by auto
	with fg gh su' s u fh have "?GE s u" by (auto simp: prc3)
      } 
      ultimately
      show ?thesis using lpo_stri_imp_nstri[of "pr" n s u] by auto
    qed
  qed
qed

lemma mpo_compat[rule_format]: 
  fixes s t u :: "('f,'v)term"
  shows "
  (snd (mpo_pr n s t) \<and> fst (mpo_pr n t u) \<longrightarrow> fst (mpo_pr n s u)) \<and>
  (fst (mpo_pr n s t) \<and> snd (mpo_pr n t u) \<longrightarrow> fst (mpo_pr n s u)) \<and>
  (snd (mpo_pr n s t) \<and> snd (mpo_pr n t u) \<longrightarrow> snd (mpo_pr n s u))" (is "?tran s t u")
proof (rule
    wf_induct[OF wf_measures[of "[\<lambda> (s,t,u). size s, \<lambda> (s,t,u). size t, \<lambda> (s,t,u). size u]"], 
    of "\<lambda> (s,t,u). ?tran s t u" "(s,t,u)", simplified], clarify)
  fix s t u :: "('f,'v)term"
  assume ind: "\<forall> s' t' u'. (size s' < size s \<longrightarrow> ?tran s' t' u') \<and> (size s' = size s \<and> (size t' < size t \<or> size t' = size t \<and> size u' < size u) \<longrightarrow> ?tran s' t' u')"
  let ?GR = "\<lambda> s t. fst (mpo_pr n s t)"
  let ?GE = "\<lambda> s t. snd (mpo_pr n s t)"
  show "?tran s t u"
  proof (cases t)
    case (Var x)
    { 
      assume "?GR t u"
      with Var have False by (cases u, auto)
    } 
    moreover
    {
      assume gr: "?GR s t" and ge: "?GE t u"
      have "?GR s u"
      proof (cases u)
	case (Var y)
	with ge `t = Var x` have "t = u" by auto
	with gr show ?thesis by auto
      next
	case (Fun h us)
	with ge Var have us: "us = []" and pri: "prl h" by auto
	from gr Var obtain f ss where s: "s = Fun f ss" by (cases s, auto)
	with gr Var obtain s' where s': "s' \<in> set ss" by (auto simp: Let_def)
	have "?GE s' u" by (simp only: Fun us, rule mpo_least_1, rule pri)
	with s' have "\<exists> s' \<in> set ss. ?GE s' u" by (auto)
	with s Fun show ?thesis by simp
      qed
    }
    moreover
    {
      assume ge1: "?GE s t" and ge2: "?GE t u"
      have "?GE s u"
      proof (cases u)
	case (Var y)
	with ge2 `t = Var x` have "t = u" by auto
	with ge1 show ?thesis by auto
      next
	case (Fun h us)
	with ge2 Var have us: "us = []" and pri: "prl h" by auto
	show ?thesis by (simp only: Fun us, rule mpo_least_1, rule pri) 
      qed
    }
    ultimately show ?thesis by blast
  next
    case (Fun g ts)
    let ?t = "Fun g ts"
    from Fun have t: "t = ?t" .
    show ?thesis 
    proof (cases s)
      case (Var x)
      {
	assume gr: "?GR s t"
	with Var Fun have False by auto
      }
      moreover
      {
	assume ge: "?GE s t" and gr: "?GR t u"
	with Var Fun have pri: "prl g" and "ts = []" by auto
	with gr Fun have False using mpo_least_2[of g n u] by auto
      }
      moreover
      {
	assume ge1: "?GE s t" and ge2: "?GE t u"
	with Var Fun have pri: "prl g" and empty: "ts = []" by auto
	have "?GE s u" by (rule mpo_least_3[of g n u], rule pri, simp add: empty[symmetric] Fun[symmetric], rule ge2)
      }
      ultimately show ?thesis by blast
    next
      case (Fun f ss)
      let ?s = "Fun f ss"
      from Fun have s: "s = ?s" .
      let ?s1 = "\<exists> s' \<in> set ss. ?GE s' t"
      let ?t1 = "\<exists> t' \<in> set ts. ?GE t' u"
      let ?ls = "length ss"
      let ?lt = "length ts"
      obtain ps pns where prc: "prc f ?ls g ?lt = (ps,pns)" by (cases "prc f ?ls g ?lt")
      let ?tran2 = "\<lambda> a b c.  
         (snd (mpo_pr n a b) \<and> fst (mpo_pr n b c) \<longrightarrow> fst (mpo_pr n a c)) \<and>
         (fst (mpo_pr n a b) \<and> snd (mpo_pr n b c) \<longrightarrow> fst (mpo_pr n a c)) \<and> 
         (snd (mpo_pr n a b) \<and> snd (mpo_pr n b c) \<longrightarrow> snd (mpo_pr n a c)) \<and> 
         (fst (mpo_pr n a b) \<and> fst (mpo_pr n b c) \<longrightarrow> fst (mpo_pr n a c))"
      from s have "\<forall> s' \<in> set ss. size s' < size s" by (auto simp: size_simps)
      with ind have ind2: "\<And> s' t' u'. \<lbrakk>s' \<in> set ss\<rbrakk> \<Longrightarrow> ?tran s' t' u'" by blast
      with mpo_stri_imp_nstri have ind3: "\<And> us s' t' u'. \<lbrakk>s' \<in> set ss; t' \<in> set ts; u' \<in> set us\<rbrakk> \<Longrightarrow> ?tran2 s' t' u'" by blast
      {
	assume ge: "?GE s t" and gr: "?GR t u"
	have "?GR s u"
	proof (cases ?s1)
	  case True
	  from this obtain s' where s': "s' \<in> set ss" and ges: "?GE s' t" by auto
	  with s s' ind2[of s' t u, simplified] ges gr have "?GR s' u" by auto
	  hence "?GE s' u" by (rule mpo_stri_imp_nstri)
	  with s' s show ?thesis by (cases u, auto simp: Let_def)
	next
	  case False
	  show ?thesis
	  proof (cases ?t1)
	    case True
	    from this obtain t' where t': "t' \<in> set ts" and ges: "?GE t' u" by auto
	    from t' t False ge s have gr2: "?GR s t'" by (cases "\<forall> t' \<in> set ts. ?GR ?s t'", auto)
	    from t' s t gr2 ges spec[OF spec[OF spec[OF ind]], of s _ u, simplified] show "?GR s u" using size_simps supt.intros by auto
	  next
	    case False
	    from t this gr obtain h us where u: "u = Fun h us" by (cases u, auto simp: Let_def)
	    let ?u = "Fun h us"	    
	    from s t u ge gr have ge: "?GE ?s ?t" and gr: "?GR ?t ?u" by auto
	    from `\<not> ?s1` t ge have st': "\<forall> t' \<in> set ts. ?GR ?s t'" by (cases ?thesis, auto)
	    from `\<not> ?t1` u gr have tu': "\<forall> u' \<in> set us. ?GR ?t u'" by (cases ?thesis, auto)
	    let ?lu = "length us"
            obtain ps2 pns2 where prc2: "prc g ?lt h ?lu = (ps2,pns2)" by (cases "prc g ?lt h ?lu")
            obtain ps3 pns3 where prc3: "prc f ?ls h ?lu = (ps3,pns3)" by (cases "prc f ?ls h ?lu")
	    from `\<not> ?s1` t ge st' have fg: "pns" by (cases ?thesis, auto simp: Let_def prc) 
	    from `\<not> ?t1` u gr tu' have gh: "pns2" by (cases ?thesis, auto simp: Let_def prc2)
            note compat = prc_compat[OF prc prc2 prc3]
            from fg gh compat have fh: "pns3" by simp 
	    from tu' s t ge spec[OF spec[OF ind], of s t, simplified] u have su': "\<forall> u' \<in> set us. ?GR ?s u'" using size_simps supt.intros by auto
	    show ?thesis
	    proof (cases ps3)
	      case True
	      with su' s u fh prc3 show ?thesis by (auto)
	    next
	      case False              
	      from False fg gh compat have nfg: "\<not> ps" and ngh: "\<not> ps2" by blast+
	      from fg `\<not> ?s1` t ge st' prc nfg have mul1: "snd (mul_ext (mpo_pr n) ss ts)" by auto
	      from gh `\<not> ?t1` u gr tu' prc2 ngh have mul2: "fst (mul_ext (mpo_pr n) ts us)" by auto
	      from mul1 mul2 mul_ext_compat[OF ind3, of ss ts us] have "fst (mul_ext (mpo_pr n) ss us)" by auto
	      with s u fh su' prc3 show ?thesis by auto
	    qed
	  qed
	qed
      }
      moreover
      {
	assume gr: "?GR s t" and ge: "?GE t u"
	have "?GR s u"
	proof (cases ?s1)
	  case True
	  from this obtain s' where s': "s' \<in> set ss" and ges: "?GE s' t" by auto
	  with s s' ind2[of s' t u, simplified] ges ge have "?GE s' u" by auto
	  with s' s show ?thesis by (cases u, auto simp: Let_def)
	next
	  case False
	  show ?thesis
	  proof (cases ?t1)
	    case True
	    from this obtain t' where t': "t' \<in> set ts" and ges: "?GE t' u" by auto
	    have tt': "?GE t t'" by (rule mpo_stri_imp_nstri[OF supt_imp_mpo_stri], auto simp: t t' supt.intros)
	    from t' t False gr s have gr2: "?GR s t'" by (cases "\<forall> t' \<in> set ts. ?GR ?s t'", auto)
	    from t' s t gr2 ges spec[OF spec[OF spec[OF ind]], of s _ u, simplified] show "?GR s u" using size_simps by auto
	  next
	    case False
	    from t this ge obtain h us where u: "u = Fun h us" by (cases u, auto simp: Let_def)
	    let ?u = "Fun h us"	    
	    from s t u ge gr have gr: "?GR ?s ?t" and ge: "?GE ?t ?u" by auto
	    from `\<not> ?s1` t gr have st': "\<forall> t' \<in> set ts. ?GR ?s t'" by (cases ?thesis, auto)
	    from `\<not> ?t1` u ge have tu': "\<forall> u' \<in> set us. ?GR ?t u'" by (cases ?thesis, auto)
	    let ?lu = "length us"
            obtain ps2 pns2 where prc2: "prc g ?lt h ?lu = (ps2,pns2)" by (cases "prc g ?lt h ?lu")
            obtain ps3 pns3 where prc3: "prc f ?ls h ?lu = (ps3,pns3)" by (cases "prc f ?ls h ?lu")
	    from `\<not> ?s1` t gr st' have fg: "pns" by (cases ?thesis, auto simp: Let_def prc) 
	    from `\<not> ?t1` u ge tu' have gh: "pns2" by (cases ?thesis, auto simp: Let_def prc2)
            note compat = prc_compat[OF prc prc2 prc3]
            from fg gh compat have fh: "pns3" by simp 
	    from tu' s t mpo_stri_imp_nstri[OF gr] spec[OF spec[OF ind], of s t, simplified] u have su': "\<forall> u' \<in> set us. ?GR ?s u'"
              using size_simps by auto
	    show ?thesis
	    proof (cases ps3)
	      case True
	      with su' s u fh prc3 show ?thesis by (auto)
	    next
	      case False
	      from False fg gh compat have nfg: "\<not> ps" and ngh: "\<not> ps2" by blast+
	      from fg `\<not> ?s1` t gr st' prc nfg have mul1: "fst (mul_ext (mpo_pr n) ss ts)"
               by auto
	      from gh `\<not> ?t1` u ge tu' prc2 ngh have mul2: "snd (mul_ext (mpo_pr n) ts us)"
               by auto
	      from mul1 mul2 mul_ext_compat[OF ind3, of ss ts us]
	      have "fst (mul_ext (mpo_pr n) ss us)" by auto
	      with s u fh su' fh prc3 show ?thesis by (auto)
	    qed
	  qed
	qed
      }
      moreover
      {
	assume ge1: "?GE s t" and ge2: "?GE t u" and ngt1: "\<not> ?GR s t" and ngt2: "\<not> ?GR t u"
	have "\<not> ?s1"
	proof
	  assume ?s1
	  with s have "?GR s t" by (cases t, auto)
	  with ngt1 show False ..
	qed
	have "\<not> ?t1"
	proof
	  assume ?t1
	  with t have "?GR t u" by (cases u, auto)
	  with ngt2 show False ..
	qed	
	from t `\<not> ?t1` ge2 obtain h us where u: "u = Fun h us" by (cases u, auto simp: Let_def)
	let ?u = "Fun h us"	    
	from s t u ge1 ge2 have ge1: "?GE ?s ?t" and ge2: "?GE ?t ?u" by auto
	from `\<not> ?s1` t ge1 have st': "\<forall> t' \<in> set ts. ?GR ?s t'" by (cases ?thesis, auto)
	from `\<not> ?t1` u ge2 have tu': "\<forall> u' \<in> set us. ?GR ?t u'" by (cases ?thesis, auto)
	let ?lu = "length us"        
        obtain ps2 pns2 where prc2: "prc g ?lt h ?lu = (ps2,pns2)" by (cases "prc g ?lt h ?lu")
        obtain ps3 pns3 where prc3: "prc f ?ls h ?lu = (ps3,pns3)" by (cases "prc f ?ls h ?lu")
	from `\<not> ?s1` t ge1 st' have fg: "pns" by (cases ?thesis, auto simp: Let_def prc) 
	from `\<not> ?t1` u ge2 tu' have gh: "pns2" by (cases ?thesis, auto simp: Let_def prc2)
        note compat = prc_compat[OF prc prc2 prc3]
        from compat fg gh have fh: pns3 by blast
	from tu' s t ge1 spec[OF spec[OF ind], of s t, simplified] u have su': "\<forall> u' \<in> set us. ?GR ?s u'" using size_simps by auto
	from `\<not> ?s1` st' ge1 ngt1 s t have nfg: "\<not> ps" by (simp add: Let_def, cases ?thesis, simp, cases ps, auto simp: prc fg) 
	from `\<not> ?t1` tu' ge2 ngt2 t u have ngh: "\<not> ps2" by (simp add: Let_def, cases ?thesis, simp, cases ps2, auto simp: prc2 gh)
	from `\<not> ?s1` st' ge1 s t fg nfg have mul1: "snd (mul_ext (mpo_pr n) ss ts)"
         by (auto simp: prc)
	from `\<not> ?t1` tu' ge2 t u gh ngh have mul2: "snd (mul_ext (mpo_pr n) ts us)"
         by (auto simp: prc2)
	from mul1 mul2 mul_ext_compat[OF ind3, of ss ts us]
	have "snd (mul_ext (mpo_pr n) ss us)" by auto
	with fg gh su' s u fh have "?GE s u" by (auto simp: prc3)
      } 
      ultimately
      show ?thesis using mpo_stri_imp_nstri[of "pr" n s u] by auto
    qed
  qed
qed


lemma supteq_imp_lpo_nstri: assumes supteq: "s \<unrhd> t" shows "snd (lpo_pr n s t)" 
proof -
  from supteq have "s = t \<or> s \<rhd> t" by auto
  thus ?thesis
  proof
    assume "s = t"
    with lpo_nstri_refl show ?thesis by auto
  next
    assume "s \<rhd> t"
    from lpo_stri_imp_nstri[OF supt_imp_lpo_stri[OF this]]
    show ?thesis .
  qed
qed

lemma supteq_imp_mpo_nstri: assumes supteq: "s \<unrhd> t" shows "snd (mpo_pr n s t)" 
proof -
  from supteq have "s = t \<or> s \<rhd> t" by auto
  thus ?thesis
  proof
    assume "s = t"
    with mpo_nstri_refl show ?thesis by auto
  next
    assume "s \<rhd> t"
    from mpo_stri_imp_nstri[OF supt_imp_mpo_stri[OF this]]
    show ?thesis .
  qed
qed

lemma lpo_trans:
  assumes one: "fst (lpo_pr n s t)"
  and two: "fst (lpo_pr n t u)"
  shows "fst (lpo_pr n s u)"
using assms lpo_compat lpo_stri_imp_nstri[OF one] by blast

lemma mpo_trans:
  assumes one: "fst (mpo_pr n s t)"
  and two: "fst (mpo_pr n t u)"
  shows "fst (mpo_pr n s u)"
using assms mpo_compat mpo_stri_imp_nstri[OF one] by blast

lemma lpo_large_sym: assumes prec: "\<And> g m. (g,m) \<in> funas_term t \<Longrightarrow> prc f (length ss) g m = (True,True)"
  and vars: "vars_term t \<subseteq> vars_term (Fun f ss)"
  shows "lpo_pr n (Fun f ss) t = (True,True)"
  using assms
proof (induct t)
  case (Var x)
  hence "x \<in> vars_term (Fun f ss)" by simp
  from supteq_Var[OF this] 
  have"Fun f ss \<rhd> Var x" unfolding subterm.le_less by simp
  from supt_imp_lpo_stri[OF this] have "fst (lpo_pr n (Fun f ss) (Var x))" .
  with lpo_stri_imp_nstri[OF this] show ?case by (cases "lpo_pr n (Fun f ss) (Var x)", auto)
next
  case (Fun g ts)
  from Fun(2) have prec: "prc f (length ss) g (length ts) = (True,True)" by simp
  { 
    fix t
    assume t: "t \<in> set ts"
    from t Fun(3) have v: "vars_term t \<subseteq> vars_term (Fun f ss)" by auto
    from t Fun(2) have p: "\<And> gg m. (gg, m) \<in> funas_term t \<Longrightarrow> prc f (length ss) gg m = (True,True)" by auto
    from Fun(1)[OF t p v] have "lpo_pr n (Fun f ss) t = (True,True)" .
  }
  thus ?case
    by (simp add: prec)
qed

lemma mpo_large_sym: assumes prec: "\<And> g m. (g,m) \<in> funas_term t \<Longrightarrow> prc f (length ss) g m = (True,True)"
  and vars: "vars_term t \<subseteq> vars_term (Fun f ss)"
  shows "mpo_pr n (Fun f ss) t = (True,True)"
  using assms
proof (induct t)
  case (Var x)
  hence "x \<in> vars_term (Fun f ss)" by simp
  from supteq_Var[OF this] 
  have"Fun f ss \<rhd> Var x" unfolding subterm.le_less by simp
  from supt_imp_mpo_stri[OF this] have "fst (mpo_pr n (Fun f ss) (Var x))" .
  with mpo_stri_imp_nstri[OF this] show ?case by (cases "mpo_pr n (Fun f ss) (Var x)", auto)
next
  case (Fun g ts)
  from Fun(2) have prec: "prc f (length ss) g (length ts) = (True,True)" by simp
  { 
    fix t
    assume t: "t \<in> set ts"
    from t Fun(3) have v: "vars_term t \<subseteq> vars_term (Fun f ss)" by auto
    from t Fun(2) have p: "\<And> gg m. (gg, m) \<in> funas_term t \<Longrightarrow> prc f (length ss) gg m = (True,True)" by auto
    from Fun(1)[OF t p v] have "mpo_pr n (Fun f ss) t = (True,True)" .
  }
  thus ?case
    by (simp add: prec)
qed

definition LPO_S_pr where "LPO_S_pr n = {(s,t) . fst (lpo_pr n s t)}"
definition LPO_NS_pr where "LPO_NS_pr n = {(s,t). snd (lpo_pr n s t)}"

lemma LPO_NS_refl: "refl (LPO_NS_pr n)"
unfolding refl_on_def LPO_NS_pr_def
 by (simp add: lpo_nstri_refl)

lemma LPO_NS_trans: "trans (LPO_NS_pr n)"
unfolding trans_def LPO_NS_pr_def 
using lpo_compat by blast

lemma LPO_S_trans: "trans (LPO_S_pr n)"
unfolding trans_def LPO_S_pr_def 
using lpo_trans by blast

lemma LPO_compat: "LPO_NS_pr n O LPO_S_pr n \<subseteq> LPO_S_pr n"
  using lpo_compat[of n] 
  by (unfold LPO_S_pr_def LPO_NS_pr_def, auto)

definition MPO_S_pr where "MPO_S_pr n = {(s,t) . fst (mpo_pr n s t)}"
definition MPO_NS_pr where "MPO_NS_pr n = {(s,t). snd (mpo_pr n s t)}"

lemma MPO_NS_refl: "refl (MPO_NS_pr n)"
unfolding refl_on_def MPO_NS_pr_def
 by (simp add: mpo_nstri_refl)

lemma MPO_NS_trans: "trans (MPO_NS_pr n)"
unfolding trans_def MPO_NS_pr_def 
using mpo_compat by blast

lemma MPO_S_trans: "trans (MPO_S_pr n)"
unfolding trans_def MPO_S_pr_def 
using mpo_trans by blast

lemma MPO_compat: "MPO_NS_pr n O MPO_S_pr n \<subseteq> MPO_S_pr n"
  using mpo_compat[of n] 
  by (unfold MPO_S_pr_def MPO_NS_pr_def, auto)

lemma LPO_S_SN: "SN (LPO_S_pr m :: ('f,'v)term ars)"
unfolding LPO_S_pr_def SN_def
proof
  fix t :: "('f,'v)term"
  let ?Rp = "\<lambda> s t. fst (lpo_pr m s t)"
  let ?R = "{(s,t). ?Rp s t}"
  let ?S = "SN_elt ?R"
  note iff = SN_elt_iff_all_reducts_SN_elt[of ?R]
  {
    fix x 
    have "?S (Var x)" unfolding iff[of "Var x"] 
    proof (intro allI impI)
      fix s
      assume "(Var x, s) \<in> ?R"
      hence False by (cases s, auto)
      thus "?S s" ..
    qed
  } note var_SN = this
  show "?S t"
  proof (induct t)
    case (Var x)
    show ?case by (rule var_SN)
  next
    case (Fun f ts)
    let ?Slist = "\<lambda> ys. \<forall> y \<in> set ys. ?S y"
    let ?cond = "\<lambda> ys xs :: ('f,'v)term list.?Slist ys \<and> fst (lex_ext (lpo_pr m) m ys xs)"
    let ?r3 = "{(ys :: ('f,'v)term list,xs). ?cond ys xs}"
    let ?r1 = "lex_two (\<lambda> (f,n) (g,m). fst (prc f n g m)) (\<lambda> (f,n) (g,m). snd (prc f n g m)) ?cond"
    let ?ge = "\<lambda> t s. snd (lpo_pr m t s)"
    let ?gr = ?Rp
    have "SN ?r3" 
    proof (rule lex_ext_SN[of "lpo_pr m" m], simp add: lpo_stri_imp_nstri)
      fix x y z 
      assume "?ge x y" and "?gr y z"
      thus "?gr x z" using lpo_compat[of m x y z] by simp
    qed
    have SN: "SN {(ab,ab'). ?r1 ab ab'}" (is "SN ?r0")
    proof (rule lex_two[OF _ _ `SN ?r3`], clarify)
      fix a1 b1 a2 b2 a3 b3
      assume ns: "snd (prc a1 b1 a2 b2)" and s: "fst (prc a2 b2 a3 b3)"
      show "fst (prc a1 b1 a3 b3)" 
        by (cases "prc a1 b1 a2 b2", cases "prc a2 b2 a3 b3", cases "prc a1 b1 a3 b3", 
          insert ns s prc_compat[of a1 b1 a2 b2 _ _ a3 b3, THEN conjunct2, THEN conjunct1], force)
    next
      let ?r'' = "{(a,a'). (\<lambda>(f,n) (g,m). fst (prc f n g m)) a a'}"
      let ?r' = "{((f, n), g, m) |f n g m. fst (prc f n g m)}"
      have "?r'' = ?r'" by auto
      thus "SN ?r''" using prc_SN by auto
    qed
    let ?m = "\<lambda> (f,ts). ((f,length ts), ts)"
    let ?r = "{(a,b).(?m a,?m b) \<in> ?r0}"
    have SN_r: "SN ?r" using SN_inv_image[OF SN, of ?m] unfolding inv_image_def by force
    let ?ind = "\<lambda> (f,ts).  ?Slist ts \<longrightarrow> ?S (Fun f ts)"
    have "?ind (f,ts)" 
    proof (rule SN_induct[OF SN_r, of ?ind])
      fix fts 
      assume ind: "\<And> gss. (fts,gss) \<in> ?r \<Longrightarrow> ?ind gss"
      show "?ind fts"
      proof (cases fts)
        case (Pair f ts)
        thus ?thesis
        proof (simp, intro impI)
	  assume ts: "?Slist ts"
	  let ?t = "Fun f ts"	
	  show "?S ?t"
	  proof (simp only: iff[of ?t], intro allI impI)
	    fix s
	    assume "(?t,s) \<in> ?R"
            hence "?gr ?t s" by simp
	    thus "?S s"
	    proof (induct s, simp add: var_SN)
	      case (Fun g ss)
	      let ?s = "Fun g ss"
	      from Fun have t_gr_s: "?gr ?t ?s" by auto
	      show "?S ?s"
	      proof (cases "\<exists> t \<in> set ts. ?ge t ?s")
	        case True
	        then obtain t where "t \<in> set ts" and ge: "?ge t ?s" by auto	      
	        with ts have "?S t" by auto
                show "?S ?s"
                proof (unfold iff[of ?s], intro allI impI)
                  fix u
                  assume "(?s,u) \<in> ?R"
                  with lpo_compat ge have u: "?gr t u" by blast
                  with `?S t`[unfolded iff[of t]]
                  show "?S u" by simp
                qed
	      next
	        case False note oFalse = this
	        let ?ls = "length ss"
	        let ?lt = "length ts"
                obtain s1 ns1 where prc1: "prc f ?lt g ?ls = (s1,ns1)" by (cases "prc f ?lt g ?ls")
	        from False t_gr_s have f_ge_g: "ns1" by (cases ?thesis, auto simp: Let_def prc1)
	        from False t_gr_s f_ge_g have small_ss: "\<forall> s \<in> set ss. ?gr ?t s" by (cases ?thesis, auto simp: Let_def)
	        with Fun have ss_S: "?Slist ss" by auto
	        show ?thesis
	        proof (cases s1)
		  case True
		  hence "((f,ts),(g,ss)) \<in> ?r"  by (simp add: prc1)
		  with ind have "?ind (g,ss)" using Pair by auto
		  with ss_S show ?thesis by auto
	        next
		  case False
		  from False oFalse t_gr_s small_ss f_ge_g have lex: "fst (lex_ext (lpo_pr m) m ts ss)" by (cases ?thesis, auto simp: Let_def prc1)
		  from False lex ts f_ge_g have "((f,ts),(g,ss)) \<in> ?r" by (simp add: prc1)
		  with ind have "?ind (g,ss)" using Pair by auto 
		  with ss_S show ?thesis by auto
	        qed
	      qed
	    qed
	  qed
        qed
      qed
    qed
    with Fun show ?case by simp
  qed
qed


lemma MPO_S_SN: "SN (MPO_S_pr m :: ('f,'v)term ars)"
unfolding MPO_S_pr_def SN_def
proof
  fix t :: "('f,'v)term"
  let ?Rp = "\<lambda> s t. fst (mpo_pr m s t)"
  let ?R = "{(s,t). ?Rp s t}"
  let ?S = "SN_elt ?R"
  note iff = SN_elt_iff_all_reducts_SN_elt[of ?R]
  {
    fix x 
    have "?S (Var x)" unfolding iff[of "Var x"] 
    proof (intro allI impI)
      fix s
      assume "(Var x, s) \<in> ?R"
      hence False by (cases s, auto)
      thus "?S s" ..
    qed
  } note var_SN = this
  show "?S t"
  proof (induct t)
    case (Var x)
    show ?case by (rule var_SN)
  next
    case (Fun f ts)
    let ?Slist = "\<lambda> ys. \<forall> y \<in> set ys. ?S y"
    let ?cond = "\<lambda> ys xs :: ('f,'v)term list.?Slist ys \<and> fst (mul_ext (mpo_pr m) ys xs)"
    let ?r3 = "{(ys :: ('f,'v)term list,xs). ?cond ys xs}"
    let ?r1 = "lex_two (\<lambda> (f,n) (g,m). fst (prc f n g m)) (\<lambda> (f,n) (g,m). snd (prc f n g m)) ?cond"
    let ?ge = "\<lambda> t s. snd (mpo_pr m t s)"
    let ?gr = ?Rp
    have "SN ?r3"
    using mpo_compat and mpo_stri_imp_nstri
     and mul_ext_SN[of "mpo_pr m", rule_format, OF mpo_nstri_refl]
     by blast
    have SN: "SN {(ab,ab'). ?r1 ab ab'}" (is "SN ?r0")
    proof (rule lex_two[OF _ _ `SN ?r3`], clarify)
      fix a1 b1 a2 b2 a3 b3
      assume ns: "snd (prc a1 b1 a2 b2)" and s: "fst (prc a2 b2 a3 b3)"
      show "fst (prc a1 b1 a3 b3)" 
        by (cases "prc a1 b1 a2 b2", cases "prc a2 b2 a3 b3", cases "prc a1 b1 a3 b3", 
          insert ns s prc_compat[of a1 b1 a2 b2 _ _ a3 b3, THEN conjunct2, THEN conjunct1], force)
    next
      let ?r'' = "{(a,a'). (\<lambda>(f,n) (g,m). fst (prc f n g m)) a a'}"
      let ?r' = "{((f, n), g, m) |f n g m. fst (prc f n g m)}"
      have "?r'' = ?r'" by auto
      thus "SN ?r''" using prc_SN by auto
    qed
    let ?m = "\<lambda> (f,ts). ((f,length ts), ts)"
    let ?r = "{(a,b).(?m a,?m b) \<in> ?r0}"
    have SN_r: "SN ?r" using SN_inv_image[OF SN, of ?m] unfolding inv_image_def by force
    let ?ind = "\<lambda> (f,ts).  ?Slist ts \<longrightarrow> ?S (Fun f ts)"
    have "?ind (f,ts)" 
    proof (rule SN_induct[OF SN_r, of ?ind])
      fix fts 
      assume ind: "\<And> gss. (fts,gss) \<in> ?r \<Longrightarrow> ?ind gss"
      show "?ind fts"
      proof (cases fts)
        case (Pair f ts)
        thus ?thesis
        proof (simp, intro impI)
	  assume ts: "?Slist ts"
	  let ?t = "Fun f ts"	
	  show "?S ?t"
	  proof (simp only: iff[of ?t], intro allI impI)
	    fix s
	    assume "(?t,s) \<in> ?R"
            hence "?gr ?t s" by simp
	    thus "?S s"
	    proof (induct s, simp add: var_SN)
	      case (Fun g ss)
	      let ?s = "Fun g ss"
	      from Fun have t_gr_s: "?gr ?t ?s" by auto
	      show "?S ?s"
	      proof (cases "\<exists> t \<in> set ts. ?ge t ?s")
	        case True
	        then obtain t where "t \<in> set ts" and ge: "?ge t ?s" by auto	      
	        with ts have "?S t" by auto
                show "?S ?s"
                proof (unfold iff[of ?s], intro allI impI)
                  fix u
                  assume "(?s,u) \<in> ?R"
                  with mpo_compat ge have u: "?gr t u" by blast
                  with `?S t`[unfolded iff[of t]]
                  show "?S u" by simp
                qed
	      next
	        case False note oFalse = this
	        let ?ls = "length ss"
	        let ?lt = "length ts"
                obtain s1 ns1 where prc1: "prc f ?lt g ?ls = (s1,ns1)" by (cases "prc f ?lt g ?ls")
	        from False t_gr_s have f_ge_g: "ns1" by (cases ?thesis, auto simp: Let_def prc1)
	        from False t_gr_s f_ge_g have small_ss: "\<forall> s \<in> set ss. ?gr ?t s" by (cases ?thesis, auto simp: Let_def)
	        with Fun have ss_S: "?Slist ss" by auto
	        show ?thesis
	        proof (cases s1)
		  case True
		  hence "((f,ts),(g,ss)) \<in> ?r"  by (simp add: prc1)
		  with ind have "?ind (g,ss)" using Pair by auto
		  with ss_S show ?thesis by auto
	        next
		  case False
		  from False oFalse t_gr_s small_ss f_ge_g have lex: "fst (mul_ext (mpo_pr m) ts ss)" by (cases ?thesis, auto simp: Let_def prc1)
		  from False lex ts f_ge_g have "((f,ts),(g,ss)) \<in> ?r" by (simp add: prc1)
		  with ind have "?ind (g,ss)" using Pair by auto 
		  with ss_S show ?thesis by auto
	        qed
	      qed
	    qed
	  qed
        qed
      qed
    qed
    with Fun show ?case by simp
  qed
qed

end


sublocale rpo_pr \<subseteq> lpo_redpair: mono_ce_redpair "LPO_S_pr n" "LPO_NS_pr n"
proof (unfold_locales, rule LPO_S_SN, rule LPO_compat)
  show "subst_closed (LPO_NS_pr n)"
    unfolding subst_closed_def 
  proof (rule subsetI2)
    fix s t :: "('f, 'v) term"
    assume "(s,t) \<in> CS (LPO_NS_pr n)"
    then obtain u v and \<sigma>::"('f, 'v) subst"
      where "(u,v) \<in> LPO_NS_pr n" and "u\<cdot>\<sigma> = s" and "v\<cdot>\<sigma> = t"
      by (cases rule: subst_closure.cases) auto
    thus "(s,t) \<in> LPO_NS_pr n" unfolding LPO_NS_pr_def using lpo_stable[of n] by auto
  qed
next
  show "subst_closed (LPO_S_pr n)"
    unfolding subst_closed_def 
  proof (rule subsetI2)
    fix s t :: "('f, 'v) term"
    assume "(s,t) \<in> CS (LPO_S_pr n)"
    then obtain u v and \<sigma> :: "('f, 'v) subst"
      where "(u,v) \<in> LPO_S_pr n" and "u\<cdot>\<sigma> = s" and "v\<cdot>\<sigma> = t"
      by (cases rule: subst_closure.cases) auto
    thus "(s,t) \<in> LPO_S_pr n" unfolding LPO_S_pr_def using lpo_stable[of n] by auto
  qed
next
  show "one_step_ctxt_closed (LPO_NS_pr n)"
    unfolding one_step_ctxt_closed_def LPO_NS_pr_def
    by (intro allI impI, simp del: lpo.simps, rule lpo_nstri_mono)
next
  show "\<exists> k. \<forall> m \<ge> k. \<forall> c d. ce_trs (c,d,m) \<subseteq> LPO_NS_pr n"
  proof (intro exI allI impI)
    fix m c d
    show "ce_trs (c,d,m) \<subseteq> LPO_NS_pr n"
      unfolding LPO_NS_pr_def 
      by (simp add: ce_trs.simps, auto simp: size_simps supt.intros lpo_stri_imp_nstri[OF supt_imp_lpo_stri])
  qed
next
  show "\<exists> k. \<forall> m \<ge> k. \<forall> c d. ce_trs (c,d,m) \<subseteq> LPO_S_pr n"
  proof (intro exI allI impI)
    fix m c d
    show "ce_trs (c,d,m) \<subseteq> LPO_S_pr n"
      unfolding LPO_S_pr_def 
      by (simp add: ce_trs.simps, auto simp: size_simps supt.intros supt_imp_lpo_stri)
  qed
qed

context rpo_pr
begin

lemma LPO_S_pr_one_step_ctxt_closed: "one_step_ctxt_closed (LPO_S_pr n)"
  unfolding one_step_ctxt_closed_def LPO_S_pr_def
  by (intro allI impI, simp del: lpo.simps, rule lpo_stri_mono)

lemma lpo_manna_ness: assumes "\<And> l r. (l,r) \<in> R \<Longrightarrow> fst (lpo_pr n l r)"
  shows "SN (rstep R)"
  by (rule manna_ness, rule LPO_S_pr_one_step_ctxt_closed[of n], insert assms, unfold LPO_S_pr_def, auto)
end

sublocale rpo_pr \<subseteq> mpo_redpair: mono_ce_redpair "MPO_S_pr n" "MPO_NS_pr n"
proof (unfold_locales, rule MPO_S_SN, rule MPO_compat)
  show "subst_closed (MPO_NS_pr n)"
    unfolding subst_closed_def 
  proof (rule subsetI2)
    fix s t :: "('f, 'v) term"
    assume "(s,t) \<in> CS (MPO_NS_pr n)"
    then obtain u v and \<sigma>::"('f, 'v) subst"
      where "(u,v) \<in> MPO_NS_pr n" and "u\<cdot>\<sigma> = s" and "v\<cdot>\<sigma> = t"
      by (cases rule: subst_closure.cases) auto
    thus "(s,t) \<in> MPO_NS_pr n" unfolding MPO_NS_pr_def using mpo_stable[of n] by auto
  qed
next
  show "subst_closed (MPO_S_pr n)"
    unfolding subst_closed_def 
  proof (rule subsetI2)
    fix s t :: "('f, 'v) term"
    assume "(s,t) \<in> CS (MPO_S_pr n)"
    then obtain u v and \<sigma> :: "('f, 'v) subst"
      where "(u,v) \<in> MPO_S_pr n" and "u\<cdot>\<sigma> = s" and "v\<cdot>\<sigma> = t"
      by (cases rule: subst_closure.cases) auto
    thus "(s,t) \<in> MPO_S_pr n" unfolding MPO_S_pr_def using mpo_stable[of n] by auto
  qed
next
  show "one_step_ctxt_closed (MPO_NS_pr n)"
    unfolding one_step_ctxt_closed_def MPO_NS_pr_def
    by (intro allI impI, simp del: mpo.simps, rule mpo_nstri_mono)
next
  show "\<exists> k. \<forall> m \<ge> k. \<forall> c d. ce_trs (c,d,m) \<subseteq> MPO_NS_pr n"
  proof (intro exI allI impI)
    fix m c d
    show "ce_trs (c,d,m) \<subseteq> MPO_NS_pr n"
      unfolding MPO_NS_pr_def 
      by (simp add: ce_trs.simps, auto simp: size_simps supt.intros mpo_stri_imp_nstri[OF supt_imp_mpo_stri])
  qed
next
  show "\<exists> k. \<forall> m \<ge> k. \<forall> c d. ce_trs (c,d,m) \<subseteq> MPO_S_pr n"
  proof (intro exI allI impI)
    fix m c d
    show "ce_trs (c,d,m) \<subseteq> MPO_S_pr n"
      unfolding MPO_S_pr_def 
      by (simp add: ce_trs.simps, auto simp: size_simps supt.intros supt_imp_mpo_stri)
  qed
qed

context rpo_pr
begin

lemma MPO_S_pr_one_step_ctxt_closed: "one_step_ctxt_closed (MPO_S_pr n)"
  unfolding one_step_ctxt_closed_def MPO_S_pr_def
  by (intro allI impI, simp del: mpo.simps, rule mpo_stri_mono)

lemma mpo_manna_ness: assumes "\<And> l r. (l,r) \<in> R \<Longrightarrow> fst (mpo_pr n l r)"
  shows "SN (rstep R)"
  by (rule manna_ness, rule MPO_S_pr_one_step_ctxt_closed[of n], insert assms, unfold MPO_S_pr_def, auto)
end

definition prc_nat :: "('f \<Rightarrow> nat \<Rightarrow> nat) \<Rightarrow> ('f \<Rightarrow> nat \<Rightarrow> 'f \<Rightarrow> nat \<Rightarrow> bool \<times> bool)"
  where "prc_nat pr \<equiv> \<lambda> f n g m. let pf = pr f n; pg = pr g m in (pg < pf, pg \<le> pf)"

definition prl_nat :: "('f \<Rightarrow> nat \<Rightarrow> nat) \<Rightarrow> 'f \<Rightarrow> bool"
  where "prl_nat pr \<equiv> \<lambda> f. pr f 0 = 0"

abbreviation pr_nat where "pr_nat pr \<equiv> (prc_nat pr, prl_nat pr)"

definition
  lpo_strict 
where
  "lpo_strict pr n \<equiv> \<lambda>(s, t). check (fst (lpo (pr_nat pr) n s t)) (''could not orient '' +#+ shows s +@+ '' >LPO '' +#+ shows t +@+ shows_nl)"

definition
  lpo_nstrict
where
  "lpo_nstrict pr n \<equiv> \<lambda> (s,t). check (snd (lpo (pr_nat pr) n s t)) (''could not orient '' +#+ shows s +@+ '' >=LPO '' +#+ shows t +@+ shows_nl)"

definition
  mpo_strict 
where
  "mpo_strict pr n \<equiv> \<lambda>(s, t). check (fst (mpo (pr_nat pr) n s t)) (''could not orient '' +#+ shows s +@+ '' >MPO '' +#+ shows t +@+ shows_nl)"

definition
  mpo_nstrict
where
  "mpo_nstrict pr n \<equiv> \<lambda> (s,t). check (snd (mpo (pr_nat pr) n s t)) (''could not orient '' +#+ shows s +@+ '' >=MPO '' +#+ shows t +@+ shows_nl)"

lemma SN_pr_nat: "SN {((f,n), g, m) | f n g m. (pr :: 'f \<Rightarrow> nat \<Rightarrow> nat) g m < pr f n}"
proof -
  have id: "{((f,n), g, m) | f n g m. pr g m < pr f n} = {(x, y). (case x of (f, n) \<Rightarrow> pr f n, case y of (f, n) \<Rightarrow> pr f n) \<in> {(x, y). y < x}}"
    by auto
  show ?thesis unfolding id
    by (rule SN_inv_image[unfolded inv_image_def, OF nat_SN.SN[simplified], of "\<lambda> (f,n). pr f n"])
qed

interpretation rpo_pr_prc: rpo_pr "prc_nat pr" "prl_nat pr"
  by (unfold_locales, unfold prc_nat_def prl_nat_def Let_def, auto simp: SN_pr_nat)

abbreviation LPO_S where "LPO_S pr n \<equiv> rpo_pr.LPO_S_pr (prc_nat pr) (prl_nat pr) n"
abbreviation LPO_NS where "LPO_NS pr n \<equiv> rpo_pr.LPO_NS_pr (prc_nat pr) (prl_nat pr) n"

interpretation LPO_redpair_impl: redpair_impl "LPO_S pr n" "LPO_NS pr n" "lpo_strict pr n" "lpo_nstrict pr n"
  by (unfold_locales, unfold lpo_strict_def lpo_nstrict_def rpo_pr_prc.LPO_S_pr_def rpo_pr_prc.LPO_NS_pr_def, auto)

interpretation LPO_ce_redpair: ce_redpair_impl "LPO_S pr n" "LPO_NS pr n" "lpo_strict pr n" "lpo_nstrict pr n" ..

interpretation LPO_ce_mono_redpair: ce_mono_redpair_impl "LPO_S pr n" "LPO_NS pr n" "lpo_strict pr n" "lpo_nstrict pr n" ch
proof(unfold_locales)
  show "one_step_ctxt_closed (LPO_S pr n)"
    unfolding one_step_ctxt_closed_def rpo_pr_prc.LPO_S_pr_def
    by (intro allI impI, simp del: lpo.simps, rule rpo_pr_prc.lpo_stri_mono)
qed

abbreviation MPO_S where "MPO_S pr n \<equiv> rpo_pr.MPO_S_pr (prc_nat pr) (prl_nat pr) n"
abbreviation MPO_NS where "MPO_NS pr n \<equiv> rpo_pr.MPO_NS_pr (prc_nat pr) (prl_nat pr) n"

interpretation MPO_redpair_impl: redpair_impl "MPO_S pr n" "MPO_NS pr n" "mpo_strict pr n" "mpo_nstrict pr n"
  by (unfold_locales, unfold mpo_strict_def mpo_nstrict_def rpo_pr_prc.MPO_S_pr_def rpo_pr_prc.MPO_NS_pr_def, auto)

interpretation MPO_ce_redpair: ce_redpair_impl "MPO_S pr n" "MPO_NS pr n" "mpo_strict pr n" "mpo_nstrict pr n" ..

interpretation MPO_ce_mono_redpair: ce_mono_redpair_impl "MPO_S pr n" "MPO_NS pr n" "mpo_strict pr n" "mpo_nstrict pr n" ch
proof(unfold_locales)
  show "one_step_ctxt_closed (MPO_S pr n)"
    unfolding one_step_ctxt_closed_def rpo_pr_prc.MPO_S_pr_def
    by (intro allI impI, simp del: mpo.simps, rule rpo_pr_prc.mpo_stri_mono)
qed

(* more efficient implementation which does not require the bound *)

function lpo_unbounded :: "('f \<Rightarrow> nat \<Rightarrow> nat) \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,'v)term \<Rightarrow> bool \<times> bool" where
  "lpo_unbounded _ (Var x) (Var y) = (False, x = y)"
| "lpo_unbounded pr (Var x) (Fun g ts) = (False, ts = [] \<and> pr g 0 = 0)"
| "lpo_unbounded pr (Fun f ss) (Var y) = (let con = list_ex (\<lambda> s. snd (lpo_unbounded pr s (Var y))) ss in (con,con))" 
| "lpo_unbounded pr (Fun f ss) (Fun g ts) = (
    if list_ex (\<lambda> s. snd (lpo_unbounded pr s (Fun g ts))) ss 
       then (True,True)
       else (let prl = pr f (length ss); prr = pr g (length ts) in 
         if prl \<ge> prr \<and> (list_all (\<lambda> t. fst (lpo_unbounded pr (Fun f ss) t)) ts)
          then if prl > prr then (True,True) 
            else lex_ext_unbounded (lpo_unbounded pr) ss ts
          else (False,False)))"
by pat_completeness auto

termination lpo_unbounded
proof (relation "measures [\<lambda> (pr,s,t). size s + size t]")
  fix "pr" and f :: 'f and ss :: "('f,'v)term list" and g :: 'f and ts :: "('f,'v)term list" and x xa xb xc i
  assume "i < length ss" and "i < length ts"
  then obtain s t where si: "s = ss ! i" and s: "s \<in> set ss" and ti: "t = ts ! i" and t: "t \<in> set ts" by auto
  from s have s1: "size s \<le> list_size size ss" by (induct ss, auto)
  from t have t1: "size t \<le> list_size size ts" by (induct ts, auto)
  from s1 t1 si ti show "((pr, ss ! i, ts ! i), pr, Fun f ss, Fun g ts) \<in> measures [\<lambda> (pr,s,t). size s + size t]"
    by auto
qed (auto simp: size_simps)

lemma lpo_to_lpo_unbounded: assumes "\<forall> f i. (f, i) \<in> funas_term s \<union> funas_term t \<longrightarrow> i \<le> n" (is "?b s t")
  shows "lpo (pr_nat pr) n s t = lpo_unbounded pr s t" (is "?e pr s t")
proof -
  let ?p = "\<lambda> u s t. ?b s t \<longrightarrow> lpo (pr_nat u) n s t = lpo_unbounded u s t"
  {
    fix p s t
    have "?p p s t"
    proof (induct rule: lpo_unbounded.induct[where P = ?p])
      case (3 p f ss y)
      show ?case 
      proof (intro impI)
        assume "?b (Fun f ss) (Var y)"
        hence "\<And> s. s \<in> set ss \<Longrightarrow> ?b s (Var y)" by auto
        with 3 have "\<And> s. s \<in> set ss \<Longrightarrow> ?e p s (Var y)" by auto
        thus "?e p (Fun f ss) (Var y)"
          by (simp add: Let_def list_ex_iff)
      qed
    next
      case (4 p f ss g ts)
      show ?case
      proof (intro impI)
        assume "?b (Fun f ss) (Fun g ts)"
        hence bs: "\<And> s. s \<in> set ss \<Longrightarrow> ?b s (Fun g ts)" and bt: "\<And> t. t \<in> set ts \<Longrightarrow> ?b (Fun f ss) t" and ss: "length ss \<le> n" and ts: "length ts \<le> n" by auto
        with 4 have s: "\<And> s. s \<in> set ss \<Longrightarrow> ?e p s (Fun g ts)" by auto
        show "?e p (Fun f ss) (Fun g ts)"
        proof (cases "\<exists> s \<in> set ss. snd (lpo (pr_nat p) n s (Fun g ts))")
          case True
          with s have "\<exists> s \<in> set ss. snd (lpo_unbounded p s (Fun g ts))" by auto
          with True show ?thesis by (auto simp: list_ex_iff)
        next
          case False note oFalse = this
          with s have False2: "\<not> list_ex (\<lambda> s. snd (lpo_unbounded p s (Fun g ts))) ss" by (auto simp: list_ex_iff)
          with bt False 4(2) have t: "\<And> t. t \<in> set ts \<Longrightarrow> ?e p (Fun f ss) t" by force
          show ?thesis 
          proof (cases "p f (length ss) \<ge> p g (length ts) \<and> (\<forall> t \<in> set ts. fst (lpo (pr_nat p) n (Fun f ss) t))")
            case False
            with t have t: "\<not> (p f (length ss) \<ge> p g (length ts) \<and> (list_all (\<lambda> t. fst (lpo_unbounded p (Fun f ss) t)) ts))" by (auto simp: list_all_iff)
            have id: "prc_nat p f (length ss) g (length ts) = (p g (length ts) < p f (length ss), p g (length ts) \<le> p f (length ss))" unfolding prc_nat_def Let_def by simp
            show ?thesis
              by (simp add: oFalse False False2 Let_def t id )
          next
            case True
            from prc_nat_def[of p] have prc_nat: "prc_nat p f = (\<lambda> n g m. let pf = p f n; pg = p g m in (pg < pf, pg \<le> pf))" by auto
            from True t have t: "p f (length ss) \<ge> p g (length ts) \<and> list_all (\<lambda> t. fst (lpo_unbounded p (Fun f ss) t)) ts" by (auto simp: list_all_iff)
            thus ?thesis proof (simp add: True False False2 Let_def t, intro conjI, simp add: prc_nat_def, intro impI,
              simp only: lex_ext_to_lex_ext_unbounded[OF ss ts, of "lpo (pr_nat p) n"], simp only: prc_nat Let_def, simp)
            assume  "p g (length ts) = p f (length ss) \<and> list_all (\<lambda> t. fst (lpo_unbounded p (Fun f ss) t)) ts"
            hence  p: "\<not> (p g (length ts) < p f (length ss))"  by simp
            {
              fix i
              assume iss: "i < length ss" and its: "i < length ts"
              with 4(3)[OF False2 _ _ t p, simplified, OF iss its]
                bs[of "ss ! i"]  bt[of "ts ! i"] have "lpo (pr_nat p) n (ss ! i) (ts ! i) = lpo_unbounded p (ss ! i) (ts ! i)" by auto
            } 
            thus "lex_ext_unbounded (lpo (pr_nat p) n) ss ts = lex_ext_unbounded (lpo_unbounded p) ss ts"
                by (simp only: lex_ext_unbounded_iff, auto )
            qed
          qed
        qed
      qed
    qed (simp_all add: prc_nat_def prl_nat_def)
  }
  with assms 
  show ?thesis by simp
qed

definition
  lpo_strict_unbounded
where
  "lpo_strict_unbounded pr \<equiv> \<lambda>(s, t). check (fst (lpo_unbounded pr s t)) (''could not orient '' +#+ shows s +@+ '' >LPO '' +#+ shows t +@+ shows_nl)"

definition
  lpo_nstrict_unbounded
where
  "lpo_nstrict_unbounded pr \<equiv> \<lambda>(s, t). check (snd (lpo_unbounded pr s t)) (''could not orient '' +#+ shows s +@+ '' >=LPO '' +#+ shows t +@+ shows_nl)"

type_synonym 'f prec_repr = "(('f \<times> nat) \<times> nat)list"


fun shows_lpo_repr :: "('f::show)prec_repr \<Rightarrow> shows"
where "shows_lpo_repr prs = (''LPO with the following precedence'' +#+ shows_nl +@+ shows_map (\<lambda>((f,n),pr).
  ''precedence('' +#+ shows f +@+ ''['' +#+ shows n +@+ '']) = '' +#+ shows pr +@+ shows_nl
) prs +@+ shows_nl
+@+ ''precedence(_) = 0'' +#+ shows_nl
)"

fun shows_mpo_repr :: "('f::show)prec_repr \<Rightarrow> shows"
where "shows_mpo_repr prs = (''MPO with the following precedence'' +#+ shows_nl +@+ shows_map (\<lambda>((f,n),pr).
  ''precedence('' +#+ shows f +@+ ''['' +#+ shows n +@+ '']) = '' +#+ shows pr +@+ shows_nl
) prs +@+ shows_nl
+@+ ''precedence(_) = 0'' +#+ shows_nl
)"


definition create_LPO_redpair :: "(('f :: show) prec_repr \<Rightarrow> ('g \<Rightarrow> nat \<Rightarrow> nat)) \<Rightarrow> 'f prec_repr \<Rightarrow> ('g :: show,'v :: show)redpair_impl"
where "create_LPO_redpair prec_repr_to_pr pr \<equiv> let p = prec_repr_to_pr pr in 
  (OK, lpo_strict_unbounded p, lpo_nstrict_unbounded p, full_af, OK, shows_lpo_repr pr)"

definition create_MPO_redpair :: "(('f :: show) prec_repr \<Rightarrow> ('g \<Rightarrow> nat \<Rightarrow> nat)) \<Rightarrow> 'f prec_repr \<Rightarrow> ('g :: show,'v :: show)redpair_impl"
where "create_MPO_redpair prec_repr_to_pr pr \<equiv> let p = prec_repr_to_pr pr in 
  (OK, mpo_strict p 0, mpo_nstrict p 0, full_af, OK, shows_mpo_repr pr)"

interpretation LPO_redpair: generic_redpair_impl "create_LPO_redpair (prec_repr_to_pr :: ('f :: show)prec_repr \<Rightarrow> ('g :: show \<Rightarrow> nat \<Rightarrow> nat))"
proof
  fix "pr" :: "('f :: show)prec_repr" and s_list ns_list :: "('g :: show,'v :: show)rule list"
  let ?rp = "create_LPO_redpair prec_repr_to_pr pr :: ('g,'v)redpair_impl"
  let ?pr = "prec_repr_to_pr pr"
  let ?af = "rp_af ?rp :: ('g af)"
  assume stri: "isOK(check_allm (rp_strict ?rp) s_list)" and nstri: "isOK (check_allm (rp_non_strict ?rp) ns_list)"
  obtain n where n: "n = max_list (map (\<lambda> (s,t). max_list (map snd ((funas_term_list s) @ (funas_term_list t)))) (s_list @ ns_list))" by auto
  {
    fix s t
    assume st: "(s,t) \<in> set (s_list @ ns_list)"
    have "\<forall> f i. (f,i) \<in> funas_term s \<union> funas_term t \<longrightarrow> i \<le> n"
    proof (intro allI impI)
      fix f i
      assume fi: "(f,i) \<in> funas_term s \<union> funas_term t"
      from max_list st have one: "max_list (map snd ((funas_term_list s) @ (funas_term_list t))) \<le> n" (is "?ml \<le> n") unfolding n by force
      from fi max_list have two: "i \<le> ?ml" by force
      from one two show "i \<le> n" by simp
    qed
    hence "lpo (pr_nat ?pr) n s t = lpo_unbounded ?pr s t" by (rule lpo_to_lpo_unbounded)
  } note main = this
  have mono: "mono_ce_af_redpair (LPO_S ?pr n) (LPO_NS ?pr n) full_af"
    by (unfold_locales, rule full_af)
  show "\<exists> S NS. ce_af_redpair S NS ?af  \<and> set s_list \<subseteq> S \<and> set ns_list \<subseteq> NS \<and> (isOK (rp_mono ?rp) \<longrightarrow> mono_ce_af_redpair S NS ?af \<and> one_step_ctxt_closed S)"
  proof (simp add: create_LPO_redpair_def Let_def, intro exI, intro conjI, 
    rule mono_ce_af_redpair.axioms(1)[OF mono], simp_all add: mono)
    show "one_step_ctxt_closed (LPO_S ?pr n :: ('g,'v)term ars)"
      by (rule LPO_ce_mono_redpair.checkMonoSound[of OK], simp)
  next
    {
      fix s t
      assume st: "(s,t) \<in> set s_list"
      with stri have "fst (lpo_unbounded ?pr s t)" by (auto simp: create_LPO_redpair_def lpo_strict_unbounded_def Let_def)
      with st have "fst (lpo (pr_nat ?pr) n s t)" by (simp add: main)
      hence "(s,t) \<in> LPO_S ?pr n" unfolding rpo_pr_prc.LPO_S_pr_def by simp
    }
    thus "set s_list \<subseteq> LPO_S ?pr n" ..
  next
    {
      fix s t
      assume st: "(s,t) \<in> set ns_list"
      with nstri have "snd (lpo_unbounded ?pr s t)" by (auto simp: create_LPO_redpair_def lpo_nstrict_unbounded_def Let_def)
      with st have "snd (lpo (pr_nat ?pr) n s t)" by (simp add: main)
      hence "(s,t) \<in> LPO_NS ?pr n" unfolding rpo_pr_prc.LPO_NS_pr_def by simp
    }
    thus "set ns_list \<subseteq> LPO_NS ?pr n" ..
  qed
qed

interpretation MPO_redpair: generic_redpair_impl "create_MPO_redpair (prec_repr_to_pr :: ('f :: show)prec_repr \<Rightarrow> ('g :: show \<Rightarrow> nat \<Rightarrow> nat))"
proof
  fix "pr" :: "('f :: show)prec_repr" and s_list ns_list :: "('g :: show,'v :: show)rule list"
  let ?rp = "create_MPO_redpair prec_repr_to_pr pr :: ('g,'v)redpair_impl"
  let ?pr = "prec_repr_to_pr pr"
  let ?af = "rp_af ?rp :: ('g af)"
  assume stri: "isOK(check_allm (rp_strict ?rp) s_list)" and nstri: "isOK (check_allm (rp_non_strict ?rp) ns_list)"
  obtain n where n: "n = (0 :: nat)" by auto
  have mono: "mono_ce_af_redpair (MPO_S ?pr n) (MPO_NS ?pr n) full_af"
    by (unfold_locales, rule full_af)
  show "\<exists> S NS. ce_af_redpair S NS ?af  \<and> set s_list \<subseteq> S \<and> set ns_list \<subseteq> NS \<and> (isOK (rp_mono ?rp) \<longrightarrow> mono_ce_af_redpair S NS ?af \<and> one_step_ctxt_closed S)"
  proof (simp add: create_MPO_redpair_def Let_def, intro exI, intro conjI, 
    rule mono_ce_af_redpair.axioms(1)[OF mono], simp_all add: mono)
    show "one_step_ctxt_closed (MPO_S ?pr n :: ('g,'v)term ars)"
      by (rule MPO_ce_mono_redpair.checkMonoSound[of OK], simp)
  next
    {
      fix s t
      assume "(s,t) \<in> set s_list"
      with stri have "fst (mpo (pr_nat ?pr) 0 s t)" by (auto simp: create_MPO_redpair_def mpo_strict_def Let_def)
      hence "(s,t) \<in> MPO_S ?pr n" unfolding n rpo_pr_prc.MPO_S_pr_def by simp
    }
    thus "set s_list \<subseteq> MPO_S ?pr n" ..
  next
    {
      fix s t
      assume "(s,t) \<in> set ns_list"
      with nstri have "snd (mpo (pr_nat ?pr) 0 s t)" by (auto simp: create_MPO_redpair_def mpo_nstrict_def Let_def)
      hence "(s,t) \<in> MPO_NS ?pr n" unfolding n rpo_pr_prc.MPO_NS_pr_def by simp
    }
    thus "set ns_list \<subseteq> MPO_NS ?pr n" ..
  qed
qed

end

