(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Tree_Automata 
imports
  Term_Impl
  Trs
  Util2
  Monadic_Check
  RBT_Map_Set_Extension
  Transitive_Closure_RBT_Impl
begin


(* syntax similar to AFP-Entry of TA *)
datatype ('q,'f)ta_rule = TA_rule 'f "'q list" 'q ("_ _ \<rightarrow> _")

record ('q,'f)ta = 
  ta_final :: "'q set"
  ta_rules :: "('q,'f) ta_rule set"
  ta_eps   :: "('q \<times> 'q)set"


fun rsym where "rsym (f qs \<rightarrow> q) = (f, length qs)"
fun rqs where "rqs (f qs \<rightarrow> q) = qs"
fun rq where "rq (f qs \<rightarrow> q) = q"


fun ta_res :: "('q,'f)ta \<Rightarrow> ('f,'q)term \<Rightarrow> 'q set"
where "ta_res TA (Var q) = {q'. (q,q') \<in> (ta_eps TA)^*}"
  | "ta_res TA (Fun f ts) = {q' | q' q qs. (f qs \<rightarrow> q) \<in> ta_rules TA \<and> (q,q') \<in> ((ta_eps TA)^*) \<and> length qs = length ts \<and> (\<forall> i < length ts. qs ! i \<in> (map (ta_res TA) ts) ! i)}"

declare ta_res.simps[code del]

definition ta_lang :: "('q,'f)ta \<Rightarrow> ('f,'q)term set"
where "ta_lang TA = {t. ground t \<and> ta_final TA \<inter> ta_res TA t \<noteq> {}}"


lemma ta_res_eps: assumes q: "q \<in> ta_res TA t" and q': "(q,q') \<in> (ta_eps TA)^*"
  shows "q' \<in> ta_res TA t"
  using q
proof (induct t)
  case (Var q'')
  from this[simplified] q' 
  show ?case by simp
next
  case (Fun f ts)
  from Fun(2)
  obtain q'' qs where rule: "f qs \<rightarrow> q'' \<in> ta_rules TA"
    and len: "length qs = length ts"
    and rec: "\<And>i. i < length ts \<Longrightarrow> qs ! i \<in> ta_res TA (ts ! i)"
    and q: "(q'',q) \<in> (ta_eps TA)^*"
    by auto
  from q q' have q': "(q'',q') \<in> (ta_eps TA)^*" by auto
  with rule len rec show ?case by auto
qed
  


(* lemma 32(i) *)
lemma ta_res_subst: assumes 
  x: "\<And> x. x \<in> vars_term t \<Longrightarrow> \<tau> x \<in> ta_res TA (\<sigma> x)"
  and q: "q \<in> ta_res TA (map_vars \<tau> t)"
  shows "q \<in> ta_res TA (t \<cdot> \<sigma>)"
using x q
proof (induct t arbitrary: q)
  case (Var x)
  from Var(1)[of x] Var(2) ta_res_eps
  show ?case by simp
next
  case (Fun f ts)
  from Fun(3)
  obtain q' qs where rule: "f qs \<rightarrow> q' \<in> ta_rules TA"
    and len: "length qs = length ts"
    and rec: "\<And>i. i < length ts \<Longrightarrow> qs ! i \<in> ta_res TA (map_vars \<tau> (ts ! i))"
    and q: "(q',q) \<in> (ta_eps TA)^*"
    by auto
  let ?ts = "map (\<lambda> t. t \<cdot> \<sigma>) ts"
  {
    fix i
    assume i: "i < length ts" 
    from i have mem: "ts ! i \<in> set ts" by auto
    hence "vars_term (ts ! i) \<subseteq> vars_term (Fun f ts)" by auto 
    with Fun(1)[OF mem Fun(2) rec[OF i]] have "qs ! i \<in> ta_res TA (ts ! i \<cdot> \<sigma>)"
      by auto
  } note main = this
  have id: "Fun f ts \<cdot> \<sigma> = Fun f ?ts" by simp
  show ?case
    by (unfold id ta_res.simps, rule, rule exI[of _ q], rule exI[of _ q'], rule exI[of _ qs], intro conjI, simp, rule rule, rule q, simp add: len, simp add: main)
qed


(* corresponds to lemma 32(ii) *)
lemma ta_res_subst_linear: 
  assumes l: "linear_term t"
  and q: "q \<in> ta_res TA (t \<cdot> \<sigma>)"
  shows "\<exists> \<tau>. q \<in> ta_res TA (map_vars \<tau> t) \<and> (\<forall> x \<in> vars_term t. \<tau> x \<in> ta_res TA (\<sigma> x))"
using l q
proof (induct t arbitrary: q)
  case (Var x)
  from Var(2) show ?case by auto
next
  case (Fun f ts)
  from Fun(3)
  obtain q' qs where rule: "f qs \<rightarrow> q' \<in> ta_rules TA"
    and len: "length qs = length ts"
    and rec: "\<And>i. i < length ts \<Longrightarrow> qs ! i \<in> ta_res TA (ts ! i \<cdot> \<sigma>)"
    and q: "(q',q) \<in> (ta_eps TA)^*"
    by auto
  {
    fix i
    assume i: "i < length ts" 
    from i have mem: "ts ! i \<in> set ts" by auto
    from Fun(2) mem have "linear_term (ts ! i)" by (auto simp: list_all_iff)
    note Fun(1)[OF mem this rec[OF i]]
  }
  hence "\<forall> i. \<exists> \<tau>. i < length ts \<longrightarrow> qs ! i \<in> ta_res TA (map_vars \<tau> (ts ! i)) \<and>
             (\<forall> x \<in> vars_term (ts ! i). \<tau> x \<in> ta_res TA (\<sigma> x))"
    by blast
  from choice[OF this] obtain \<tau> where
    ind: "\<And> i. i < length ts \<Longrightarrow> qs ! i \<in> ta_res TA (map_vars (\<tau> i) (ts ! i)) \<and> 
             (\<forall> x \<in> vars_term (ts ! i). (\<tau> i) x \<in> ta_res TA (\<sigma> x))"
    by auto
  from Fun(2) have "is_partition (map vars_term ts)" by auto
  from subst_merge[OF this, of \<tau>] obtain \<tau>2 where \<tau>2: "\<And> i. i < length ts \<Longrightarrow> \<forall> x \<in> vars_term (ts ! i). \<tau>2 x = \<tau> i x" by auto
  show ?case
  proof (rule exI[of _ \<tau>2], intro conjI)
    show "\<forall> x \<in> vars_term (Fun f ts). \<tau>2 x \<in> ta_res TA (\<sigma> x)"
    proof
      fix x
      assume "x \<in> vars_term (Fun f ts)"
      then obtain t where "t \<in> set ts" and "x \<in> vars_term t" by auto
      then obtain i where i: "i < length ts" and x: "x \<in> vars_term (ts ! i)" unfolding set_conv_nth by auto
      from ind[OF i] \<tau>2[OF i] x show "\<tau>2 x \<in> ta_res TA (\<sigma> x)" by auto
    qed
  next
    { 
      fix i
      assume i: "i < length ts"
      have id: "map_vars \<tau>2 (ts ! i) = map_vars (\<tau> i) (ts ! i)" 
        by (rule map_vars_vars_term, insert \<tau>2[OF i], auto)
      from ind[OF i, THEN conjunct1] 
      have "qs ! i \<in> ta_res TA (map_vars \<tau>2 (ts ! i))" unfolding id .
    } note main = this
    have "q \<in> ta_res TA (Fun f (map (map_vars \<tau>2) ts))"
      by (unfold ta_res.simps, rule, rule exI[of _ q], rule exI[of _ q'], rule exI[of _ qs], intro conjI, simp, rule rule, rule q, simp add: len, simp add: main)
    thus "q \<in> ta_res TA (map_vars \<tau>2 (Fun f ts))" by auto
  qed
qed

definition ta_rqs :: "('q,'f)ta \<Rightarrow> 'q set"
where "ta_rqs TA \<equiv> {q'. \<exists> q. (q \<in> rq ` ta_rules TA) \<and> (q,q') \<in> (ta_eps TA)^*}"

lemma ta_rqs_res: assumes ground: "ground t" 
  shows "ta_res TA t \<subseteq> ta_rqs TA"
proof
  fix q 
  assume q: "q \<in> ta_res TA t"
  from ground obtain f ts where t: "t = Fun f ts" by (cases t, auto)
  from q[unfolded t] obtain q' qs where "(f qs \<rightarrow> q') \<in> ta_rules TA" 
    and q: "(q',q) \<in> (ta_eps TA)^*" by auto
  thus "q \<in> ta_rqs TA" unfolding ta_rqs_def by force
qed    


definition ta_syms :: "('q,'f)ta \<Rightarrow> 'f sig"
where "ta_syms TA \<equiv> rsym ` ta_rules TA"

lemma ta_syms_res: "q \<in> ta_res TA t \<Longrightarrow> t \<in> wf_terms (ta_syms TA)"
proof (induct t arbitrary: q)
  case (Var x)
  show ?case ..
next
  case (Fun f ts)
  from Fun(2)
  obtain q' qs where rule: "f qs \<rightarrow> q' \<in> ta_rules TA"
    and len: "length qs = length ts"
    and rec: "\<forall> i < length ts. qs ! i \<in> ta_res TA (ts ! i)"
    by auto
  {
    fix t 
    assume mem: "t \<in> set ts"
    then obtain i where i: "i < length ts"
      and t: "t = ts ! i" unfolding set_conv_nth by auto
    from Fun(1)[OF mem[unfolded t] rec[THEN spec, THEN mp[OF _ i]]]
    have "t \<in> wf_terms (ta_syms TA)" unfolding t by auto
  } note rec = this
  from rule len have f: "(f,length ts) \<in> ta_syms TA" unfolding ta_syms_def
    by force
  show ?case
    by (rule, rule f, rule rec)
qed

lemma ta_syms_lang: "ta_lang TA \<subseteq> wf_terms (ta_syms TA)"
proof
  fix t
  assume "t \<in> ta_lang TA"
  then obtain q where "q \<in> ta_res TA t" unfolding ta_lang_def
    by auto
  from ta_syms_res[OF this] show "t \<in> wf_terms (ta_syms TA)" .
qed


(* the restrictions on vars_term and \<tau> are made to ensure that we will only
   have to test finitely many \<tau> *)
definition ta_rule_compatible :: "('q,'f)ta \<Rightarrow> ('f,'v)rule \<Rightarrow> bool"
where "ta_rule_compatible TA \<equiv> \<lambda> (l,r). (\<forall> \<tau>. (\<tau> ` vars_term l \<subseteq> ta_rqs TA) \<longrightarrow> 
  ta_res TA (map_vars \<tau> l) \<subseteq> ta_res TA (map_vars \<tau> r))"

  
definition ta_trs_compatible :: "('q,'f)ta \<Rightarrow> ('f,'v)trs \<Rightarrow> bool"
where "ta_trs_compatible TA R \<equiv> \<forall> l r. (l,r) \<in> R \<and> funas_term l \<subseteq> ta_syms TA \<longrightarrow> ta_rule_compatible TA (l,r)" 


lemma ta_trs_compatible_res: assumes 
      comp: "ta_trs_compatible TA R"
  and ll: "left_linear_trs R"
  and wf_trs: "\<And> l r. (l,r) \<in> R \<Longrightarrow> vars_term r \<subseteq> vars_term l"
  and step: "(s,t) \<in> rstep R"
  and ground: "ground s"
  shows "ta_res TA s \<subseteq> ta_res TA t"
  using step ground
proof (induct)
  case (IH C \<sigma> l r)
  hence ground: "ground (l \<cdot> \<sigma>)" by simp
  show ?case
  proof (induct C)
    case (More f bef C aft)
    let ?args = "\<lambda> l. bef @ C\<langle>l \<cdot> \<sigma>\<rangle> # aft"
    let ?l = "Fun f (?args l)"
    let ?r = "Fun f (?args r)"
    let ?n = "Suc (length bef + length aft)"
    let ?m = "length bef"
    show ?case
    proof
      fix q
      assume "q \<in> ta_res TA (More f bef C aft)\<langle>l\<cdot>\<sigma>\<rangle>"
      hence "q \<in> ta_res TA ?l" by simp
      from this[unfolded ta_res.simps] obtain q' qs where 
        rule: "f qs \<rightarrow> q' \<in> ta_rules TA"
        and len: "length qs = ?n"
        and rec: "\<forall>i < ?n. qs ! i \<in> map (ta_res TA) (?args l) ! i"
        and q: "(q',q) \<in> (ta_eps TA)^*"
        by auto
      have len2: "\<And>t. ?m < length (?args t)" by auto
      from rec[THEN spec, of ?m, unfolded nth_map[OF len2]]
      have "qs ! ?m \<in> ta_res TA (C\<langle>l \<cdot> \<sigma>\<rangle>)"  by auto
      with More have m: "qs ! ?m \<in> ta_res TA (C\<langle>r\<cdot>\<sigma>\<rangle>)" by auto
      { 
        fix i
        assume i: "i < ?n"
        have "qs ! i \<in> map (ta_res TA) (?args r) ! i" 
        proof (cases "i = ?m")
          case True
          show ?thesis using m unfolding True
            by (simp add: nth_append)
        next
          case False
          note rec = rec[THEN spec, THEN mp, OF i]
          show ?thesis 
          proof (cases "i < ?m")
            case True
            show ?thesis using rec
              by (simp add: nth_append True)
          next
            case False
            with `i \<noteq> ?m` have "i - ?m > 0" by auto
            then obtain j where im: "i - ?m = Suc j" by (cases "i - ?m", auto)
            show ?thesis using rec 
              by (simp add: nth_append False im)
          qed
        qed
      }
      hence rec: "\<forall>i < ?n. qs ! i \<in> map (ta_res TA) (?args r) ! i"
        by simp        
      have "q \<in> ta_res TA ?r" 
        unfolding ta_res.simps
        by (rule, rule exI[of _ q], rule exI[of _ q'], rule exI[of _ qs],
          insert rule len rec q, simp)
      thus "q \<in> ta_res TA (More f bef C aft)\<langle>r\<cdot>\<sigma>\<rangle>" by simp
    qed
  next
    case Hole
    show ?case
    proof (unfold ctxt_apply.simps, rule)
      fix q
      assume q: "q \<in> ta_res TA (l \<cdot> \<sigma>)"
      from ll IH(1) have "linear_term l" unfolding left_linear_trs_def by auto
      from ta_res_subst_linear[OF this q] obtain \<tau> where
        q: "q \<in> ta_res TA (map_vars \<tau> l)" and 
        vars: "\<forall> x \<in> vars_term l. \<tau> x \<in> ta_res TA (\<sigma> x)" by auto
      from in_funas_term_wf_terms[OF _ ta_syms_res[OF q]] 
      have "funas_term l \<subseteq> ta_syms TA" by auto
      with comp[unfolded ta_trs_compatible_def] IH(1) have
        comp: "ta_rule_compatible TA (l,r)" by auto
      have "\<tau> ` vars_term l \<subseteq> ta_rqs TA" 
      proof 
        fix q
        assume "q \<in> \<tau> ` vars_term l"
        then obtain x where x: "x \<in> vars_term l" and q: "q = \<tau> x" by auto
        with vars have q: "q \<in> ta_res TA (\<sigma> x)" by auto
        from ground x have "ground (\<sigma> x)" by auto
        from ta_rqs_res[OF this] q show "q \<in> ta_rqs TA" by auto
      qed
      with comp[unfolded ta_rule_compatible_def] q
      have q: "q \<in> ta_res TA (map_vars \<tau> r)" by auto
      from ta_res_subst[OF _ q] wf_trs[OF IH(1)] vars
      show "q \<in> ta_res TA (r \<cdot> \<sigma>)" by force
    qed
  qed
qed


lemma ta_trs_compatible_lang: assumes 
      comp: "ta_trs_compatible TA R"
  and ll: "left_linear_trs R"
  and wf_trs: "\<And> l r. (l,r) \<in> R \<Longrightarrow> vars_term r \<subseteq> vars_term l"
  and step: "(s,t) \<in> rstep R"
  and s: "s \<in> ta_lang TA"
  shows "t \<in> ta_lang TA"
proof -
  from s[unfolded ta_lang_def]
  have g: "ground s" and acc: "ta_final TA \<inter> ta_res TA s \<noteq> {}" by auto
  from rstep_ground[OF wf_trs g step] have gt: "ground t" .
  from ta_trs_compatible_res[OF comp ll wf_trs step g] acc 
  have acc: "ta_final TA \<inter> ta_res TA t \<noteq> {}" by auto
  from gt acc show "t \<in> ta_lang TA" unfolding ta_lang_def by auto
qed


(* lemma 33 *)
lemma ta_trs_compatible: assumes 
      comp: "ta_trs_compatible TA R"
  and ll: "left_linear_trs R"
  and wf_trs: "\<And> l r. (l,r) \<in> R \<Longrightarrow> vars_term r \<subseteq> vars_term l"
  and L: "L \<subseteq> ta_lang TA"
  shows "{t | s t. s \<in> L \<and> (s,t) \<in> (rstep R)^*} \<subseteq> ta_lang TA"
proof -
  {
    fix s t
    assume "(s,t) \<in> (rstep R)^*" and "s \<in> L" 
    hence "t \<in> ta_lang TA"
    proof (induct)
      case base
      thus ?case using L by auto
    next
      case (step t u)
      hence t: "t \<in> ta_lang TA" by simp
      show ?case
        by (rule ta_trs_compatible_lang[OF comp ll wf_trs step(2) t])
    qed
  }
  thus ?thesis by auto
qed

lemma finite_ta_syms: assumes "finite (ta_rules TA)" shows "finite (ta_syms TA)"
  unfolding ta_syms_def using assms by auto

    

(* lemma 35 *)
lemma ta_compact: assumes finite: "finite (ta_rules TA)"
  and comp: "ta_trs_compatible TA R"
  and ll: "left_linear_trs R"
  and wf_trs: "\<And> l r. (l,r) \<in> R \<Longrightarrow> vars_term r \<subseteq> vars_term l"
  and L: "L \<subseteq> ta_lang TA"
  shows "compact L R"
  unfolding compact_def
  by (rule exI[of _ "ta_syms TA"], intro conjI[OF finite_ta_syms[OF finite]], rule subset_trans[OF ta_trs_compatible[OF comp ll wf_trs L] ta_syms_lang])

definition ta_contains :: "('f \<times> nat)set \<Rightarrow> ('q,'f)ta \<Rightarrow> 'q set \<Rightarrow> bool"
  where "ta_contains F TA Q \<equiv> Q \<subseteq> ta_final TA
  \<and> (\<forall> f qs. (f,length qs) \<in> F \<and> set qs \<subseteq> Q \<longrightarrow> (\<exists> q q'. (f qs \<rightarrow> q) \<in> ta_rules TA \<and> q' \<in> Q \<and> (q,q') \<in> (ta_eps TA)^*))"

lemma ta_contains: 
  assumes contain: "ta_contains F TA qs"
  shows "(wf_terms F :: ('f,'q)term set) \<inter> {t. ground t} \<subseteq> ta_lang TA"
proof (clarify)
  fix t :: "('f,'q)term"
  note contain = contain[unfolded ta_contains_def, simplified]
  assume "t \<in> wf_terms F" and g: "ground t"
  hence "\<exists> q \<in> qs. q \<in> ta_res TA t"
  proof (induct)
    case (Var x) thus ?case by simp
  next
    case (Fun f ts)
    {
      fix i
      assume "i < length ts"
      hence t: "ts ! i \<in> set ts" by auto
      from Fun(4) t have gt: "ground (ts ! i)" by auto
      note Fun(3)[OF t gt] 
    } hence "\<forall> i. \<exists> q. i < length ts \<longrightarrow> q \<in> qs \<and> q \<in> ta_res TA (ts ! i)" by auto
    from choice[OF this] obtain qs' where qs'res: "\<And>i. i < length ts \<Longrightarrow> qs' i \<in> qs \<and> qs' i \<in> ta_res TA (ts ! i)" by auto
    let ?qs = "map qs' [0 ..< length ts]"
    have qs': "\<And> i. i < length ts \<Longrightarrow>  ?qs ! i = qs' i" by auto
    with qs'res have qs'res: "\<forall> i < length ts. ?qs ! i \<in> qs \<and> ?qs ! i \<in> ta_res TA (ts ! i)" by auto
    from contain[THEN conjunct2, THEN spec[of _ f], THEN spec[of _ ?qs]] Fun(1) qs'res 
    obtain q' q where q: "q \<in> qs" and q': "(q',q) \<in> (ta_eps TA)^*" and rule: "f ?qs \<rightarrow> q' \<in> ta_rules TA" by auto
    show ?case unfolding ta_res.simps
      by (rule bexI[OF _ q], rule, rule exI[of _ q], rule exI[of _ q'], rule exI[of _ ?qs], insert rule qs'res, simp add: q')
  qed
  then obtain q where q1: "q \<in> qs" and q2: "q \<in> ta_res TA t" by auto
  from contain[THEN conjunct1] q1 have q1: "q \<in> ta_final TA"
    by auto
  show "t \<in> ta_lang TA" unfolding ta_lang_def using q1 q2 g by auto
qed  



text {* here the part about tree automata implementation starts *}

datatype ('q,'f)ta_rule_impl = TA_rule_impl 'f "'q list" 'q "'q rs" 

fun ta_rule_conv :: "('q,'f)ta_rule_impl \<Rightarrow> ('q,'f)ta_rule" 
  where "ta_rule_conv (TA_rule_impl f qs q qs') = (TA_rule f qs q)"

fun conv_ta_rule :: "('q \<Rightarrow> 'q rs) \<Rightarrow> ('q,'f)ta_rule \<Rightarrow> ('q,'f)ta_rule_impl" 
  where "conv_ta_rule eps (TA_rule f qs q) = (TA_rule_impl f qs q (eps q))"

record ('q,'f)ta_impl =
  ta_final_impl :: "'q rs"
  ta_rules_impl :: "('f \<times> nat, ('q,'f) ta_rule_impl list) rm"
  ta_rqs_impl :: "'q list"
  ta_rqs_set :: "'q rs"
  ta_eps_impl :: "'q \<Rightarrow> 'q rs"


definition ta_of :: "('q :: linorder,'f :: linorder)ta_impl \<Rightarrow> ('q,'f)ta"
  where "ta_of TA = \<lparr> ta_final = set (rs_to_list (ta_final_impl TA)), 
  ta_rules = \<Union>{ta_rule_conv ` (set (rm_set_lookup (ta_rules_impl TA) fn)) | fn. True},
  ta_eps = (\<lambda> (q,q'). q' \<in> rs_\<alpha> (ta_eps_impl TA q)) \<rparr>"

fun rqs_impl where "rqs_impl (TA_rule_impl f qs q qs') = qs"
fun rq_impl where "rq_impl (TA_rule_impl f qs q qs') = qs'"
fun rsym_impl where "rsym_impl (TA_rule_impl f qs q qs') = (f,length qs)"

fun ta_res_list :: "('f \<times> nat,('q :: linorder,'f :: linorder)ta_rule_impl list)rm \<Rightarrow> ('q \<Rightarrow> 'q rs) \<Rightarrow> ('f,'q)term \<Rightarrow> 'q rs"
where "ta_res_list TA eps (Var q) = eps q"
  | "ta_res_list TA eps (Fun f ts) = (
        let rec = map (ta_res_list TA eps) ts;
            n = length ts;
            rules = rm_set_lookup TA (f,n);
            arules = filter (\<lambda> rule. let qs = rqs_impl rule in list_all (\<lambda> i. rs_memb (qs ! i) (rec ! i)) [0 ..< n]) rules;
            qs = map (\<lambda> rule. rq_impl rule) arules   
           in rs_Union qs)"



locale ta_inv = rm_set "ta_rules_impl TA" rsym_impl "conv_ta_rule (ta_eps_impl TA) ` ta_rules (ta_of TA)" for
  TA :: "('q :: linorder,'f :: linorder)ta_impl" +
  assumes ta_rqs_set: "rs_\<alpha> (ta_rqs_set TA) = ta_rqs (ta_of TA)"
  and     ta_rqs_impl: "set (ta_rqs_impl TA) = ta_rqs (ta_of TA)"
  and     ta_eps: "\<And> q. rs_\<alpha> (ta_eps_impl TA q) = {q'. (q,q') \<in> (ta_eps (ta_of TA))^*}"
begin
abbreviation ta_rm_set_lookup where "ta_rm_set_lookup \<equiv> rm_set_lookup (ta_rules_impl TA)"
abbreviation TA' :: "('q,'f)ta" where "TA' \<equiv> ta_of TA"

lemma rm_set_lookup1: assumes "rule \<in> set (ta_rm_set_lookup (f,n))"
  shows "\<exists> qs q qs'. rule = (TA_rule_impl f qs q qs') \<and> n = length qs \<and> rs_\<alpha> qs' = {q'. (q,q') \<in> (ta_eps TA')^*} \<and> (f qs \<rightarrow> q) \<in> ta_rules TA'"
proof -
  from assms have "rsym_impl rule = (f,n)" and mem: "rule \<in> conv_ta_rule (ta_eps_impl TA) ` ta_rules TA'"
    using assms unfolding rm_set_lookup by auto
  then obtain qs q qs' where n: "n = length qs" and rule: "rule = TA_rule_impl f qs q qs'" by (cases rule, auto)
  from mem obtain rrule where id: "rule = conv_ta_rule (ta_eps_impl TA) rrule" and rrule: "rrule \<in> ta_rules TA'" by auto
  from id[unfolded rule] have idd: "rrule = (f qs \<rightarrow> q)" by (cases rrule, auto) 
  from id[unfolded rule idd] have "rs_\<alpha> qs' = rs_\<alpha> (ta_eps_impl TA q)" by auto
  from this[unfolded ta_eps] rrule
  show ?thesis unfolding idd rule n
    by auto
qed

lemma rm_set_lookup2: assumes "rule \<in> conv_ta_rule (ta_eps_impl TA) ` (ta_rules TA')"
  shows "rule \<in> set (ta_rm_set_lookup (rsym_impl rule))"
  using assms unfolding rm_set_lookup
  by (cases rule, auto)

lemma rm_set_lookup3: assumes "rule \<in> conv_ta_rule (ta_eps_impl TA) ` (ta_rules TA')"
  shows "\<exists> rules. Some rules = rm_\<alpha> (ta_rules_impl TA) (rsym_impl rule) \<and> rule \<in> set rules"
  using rm_set_lookup2[OF assms] by (auto simp: rm_set_lookup_def)

(*
lemma ta_eps_impl[simp]: "rs_\<alpha> (ta_eps_impl TA q) = ta_eps TA' q"
  unfolding ta_of_def by auto
*)

lemma ta_res_list: 
  shows "rs_\<alpha> (ta_res_list (ta_rules_impl TA) (ta_eps_impl TA) t) = ta_res TA' t"
proof (induct t)
  case (Var x) thus ?case by (simp add: ta_eps)
next
  case (Fun f ts)
  show ?case
  proof (rule set_eqI)
    fix q
    show "(q \<in> rs_\<alpha> (ta_res_list (ta_rules_impl TA) (ta_eps_impl TA) (Fun f ts))) = (q \<in> ta_res (ta_of TA) (Fun f ts))"
      (is "(q \<in> ?L) = (q \<in> ?S)")
    proof
      assume "q \<in> ?L"
      then obtain qs' rule where mem: "rule \<in> set (ta_rm_set_lookup (f,length ts))"
        and q: "rq_impl rule = qs'"
        and q': "q \<in> rs_\<alpha> qs'"
        and args: "\<forall> i \<in> {0 ..< length ts}.
             rs_memb (rqs_impl rule ! i) (map (ta_res_list (ta_rules_impl TA) (ta_eps_impl TA)) ts ! i)"
        by (auto simp: Let_def list_all_iff rs_correct)
      from rm_set_lookup1[OF mem] 
      obtain qs qs'' q' where id: "rule = TA_rule_impl f qs q' qs''" and
        len: "length ts = length qs" and 
        qs'': "rs_\<alpha> qs'' = {q. (q',q) \<in> (ta_eps TA')^*}" and
        rule: "f qs \<rightarrow> q' \<in> ta_rules TA'" by auto
      from q id have qs': "qs' = qs''" by auto
      from q'[unfolded qs' qs''] have qq': "(q',q) \<in> (ta_eps TA')^*" by simp
      show "q \<in> ?S"
      proof (unfold ta_res.simps, rule, rule exI[of _ q], rule exI[of _ q'], rule exI[of _ qs], intro conjI, simp, simp add: rule, rule qq', simp add: len, intro allI impI)
        fix i
        assume i: "i < length ts"
        hence mem: "ts ! i \<in> set ts" by auto
        from args[unfolded id, simplified] i len
        have "rs_memb (qs ! i) (ta_res_list (ta_rules_impl TA) (ta_eps_impl TA) (ts ! i))" by auto
        with Fun[OF mem] i 
        show "qs ! i \<in> map (ta_res TA') ts ! i" by (simp add: rs_correct)
      qed
    next
      assume "q \<in> ?S"
      then obtain q' qs where rule: "f qs \<rightarrow> q' \<in> ta_rules TA'"
        and len: "length qs = length ts"
        and args: "\<forall> i < length ts. qs ! i \<in> map (ta_res TA') ts ! i"
        and q: "(q',q) \<in> (ta_eps TA')^*"
        by auto
      let ?rule = "TA_rule_impl f qs q' (ta_eps_impl TA q')"
      have rrule: "?rule = conv_ta_rule (ta_eps_impl TA) (f qs \<rightarrow> q')" by auto
      have rrule: "?rule \<in> conv_ta_rule (ta_eps_impl TA) ` ta_rules TA'"
        unfolding rrule using rule by blast
      have rule: "?rule \<in> set (ta_rm_set_lookup (f,length qs))"
        using rm_set_lookup3[OF rrule]
        unfolding rm_set_lookup_def by auto
      {
                fix i
        assume i: "i < length ts"
        hence mem: "ts ! i \<in> set ts" by simp
        have "qs ! i \<in> rs_\<alpha> (ta_res_list (ta_rules_impl TA) (ta_eps_impl TA) (ts ! i))"
          unfolding Fun[OF mem] using args i by auto
      } note args = this
      show "q \<in> ?L"
        by (simp add: Let_def list_all_iff rs_correct, intro exI conjI, rule rule[unfolded len], insert args q, simp add: len, simp add: ta_eps)
    qed
  qed
qed
end

(* the following performs a brute force enumeration over
  all possible substitutions; this might be a bottle-neck and can be improved *)
definition ta_rule_compatible_list :: "('q :: linorder,'f :: linorder)ta_impl \<Rightarrow> ('f,'v)rule \<times> 'v list \<Rightarrow> (('f,'q)rule \<times> 'q) check"
where "ta_rule_compatible_list TA \<equiv> let rm = ta_rules_impl TA; eps = ta_eps_impl TA in (\<lambda> ((l,r),xs). 
          (check_allm (\<lambda> vec. let \<tau> = assign_from_vector vec;
                               l\<tau> = map_vars \<tau> l;
                               r\<tau> = map_vars \<tau> r;
                               rhs = rs_\<alpha> (ta_res_list rm eps r\<tau>)
                            in  (check_allm (\<lambda> q. check (q \<in> rhs) ((l\<tau>,r\<tau>),q)) (rs_to_list (ta_res_list rm eps l\<tau>))))
                        (enum_vectors (ta_rqs_impl TA) xs)))"


context ta_inv
begin

(* I believe that also equivalence holds but was to lazy to prove it *)
lemma ta_rule_compatible_list: assumes wf: "vars_term r \<subseteq> vars_term l"
  and comp: "isOK(ta_rule_compatible_list TA ((l,r),vars_term_list l))"
  shows "ta_rule_compatible TA' (l,r)"
proof (cases "ta_rqs TA' = {}")
  case False
  show ?thesis
    unfolding ta_rule_compatible_def 
  proof(clarify)
    fix \<tau> q
    assume q: "q \<in> ta_res TA' (map_vars \<tau> l)"
      and \<tau>: "\<tau> ` vars_term l \<subseteq> ta_rqs TA'"
    let ?Q = "ta_rqs_impl TA"
    from False have "?Q \<noteq> []" using ta_rqs_impl by auto
    from enum_vectors_complete[OF this, of "vars_term_list l"]
    obtain vec where vec: "vec \<in> set (enum_vectors ?Q (vars_term_list l))" and \<tau>2: "\<And> x c. x \<in> set (vars_term_list l) \<Longrightarrow> c \<in> set ?Q \<Longrightarrow>  \<tau> x = c \<longrightarrow> assign_from_vector vec x = c" by blast
    let ?\<tau> = "assign_from_vector vec"
    {
      fix x
      assume x: "x \<in> vars_term l"
      with \<tau> have "\<tau> x \<in> ta_rqs TA'" by auto    
      from \<tau>2[unfolded vars_term_list_sound ta_rqs_impl, OF x this] have "\<tau> x = ?\<tau> x"
        by simp
    } note \<tau>2 = this
    have l: "map_vars \<tau> l = map_vars ?\<tau> l"
      by (rule map_vars_vars_term[OF \<tau>2])
    have r: "map_vars \<tau> r = map_vars ?\<tau> r"
      by (rule map_vars_vars_term, insert \<tau>2 wf, auto)
    from comp[unfolded ta_rule_compatible_list_def Let_def, simplified, THEN bspec[OF _ vec]] q 
    show "q \<in> ta_res TA' (map_vars \<tau> r)" unfolding l r by (auto simp: rs_correct ta_res_list)
  qed
next
  case True
  show "ta_rule_compatible TA' (l,r)"
    unfolding ta_rule_compatible_def
  proof (clarify)
    fix \<tau> q
    assume q: "q \<in> ta_res TA' (map_vars \<tau> l)"
      and \<tau>: "\<tau> ` vars_term l \<subseteq> ta_rqs TA'"    
    show "q \<in> ta_res TA' (map_vars \<tau> r)"
    proof (cases l)
      case (Fun f ls)
      from q[unfolded Fun] True
      show ?thesis unfolding ta_rqs_def by auto
    next
      case (Var x)
      from \<tau>[unfolded Var ta_rqs_def]  True
      show ?thesis unfolding ta_rqs_def by simp
    qed
  qed    
qed
end


fun ta_res_list_all :: "'q rs \<Rightarrow> ('f \<times> nat,('q :: linorder,'f :: linorder)ta_rule_impl list)rm \<Rightarrow> ('f,'v)term \<Rightarrow> 'q rs"
where "ta_res_list_all Q TA (Var _) = Q"
  | "ta_res_list_all Q TA (Fun f ts) = (
        let rec = map (ta_res_list_all Q TA) ts;
            n = length ts;
            rules = rm_set_lookup TA (f,n);
            arules = filter (\<lambda> rule. let qs = rqs_impl rule in list_all (\<lambda> i. rs_memb (qs ! i) (rec ! i)) [0 ..< n]) rules;
            qs = map (\<lambda> rule. rq_impl rule) arules   
           in rs_Union qs)"

context ta_inv
begin
lemma ta_res_list_all: assumes "\<tau> ` vars_term t \<subseteq> ta_rqs TA'"
  shows "rs_\<alpha> (ta_res_list (ta_rules_impl TA) (ta_eps_impl TA) (map_vars \<tau> t)) \<subseteq> rs_\<alpha> (ta_res_list_all (ta_rqs_set TA) (ta_rules_impl TA) t)"
  using assms
proof (induct t)
  case (Var q)
  with assms  have q: "\<tau> q \<in> ta_rqs TA'" by auto
  { 
    fix q'
    assume q': "(\<tau> q, q') \<in> (ta_eps TA')^*"
    from q q' 
    have "q' \<in> ta_rqs TA'" unfolding ta_rqs_def
      by auto
  }
  thus ?case by (auto simp: ta_eps ta_rqs_set)
next
  case (Fun f ts)
  {
    fix rule 
    assume 
      args: "\<forall> i \<in> {0..<length ts}. rs_memb (rqs_impl rule ! i) (map (ta_res_list (ta_rules_impl TA) (ta_eps_impl TA)) (map (map_vars \<tau>) ts) ! i)"
    have args: "\<forall> i \<in> {0..<length ts}. rs_memb (rqs_impl rule ! i)  (map (ta_res_list_all (ta_rqs_set TA) (ta_rules_impl TA)) ts ! i)" (is "\<forall> i \<in> _. ?c i")
    proof -
      {
        fix i
        assume i: "i < length ts"
        from i have mem: "ts ! i \<in> set ts" by auto
        with Fun(2) have "\<tau> ` vars_term (ts ! i) \<subseteq> ta_rqs TA'" by auto
        from Fun(1)[OF mem this] args i
        have "?c i" by (auto simp: rs_correct)
      }
      thus ?thesis by auto
    qed
  }
  thus ?case by (auto simp: Let_def rs_correct list_all_iff)
qed
end

definition ta_rule_compatible_heuristic :: "('q :: linorder,'f :: linorder)ta_impl \<Rightarrow> ('f,'v)term \<Rightarrow> bool" where
"ta_rule_compatible_heuristic TA l \<equiv> rs_isEmpty (ta_res_list_all (ta_rqs_set TA) (ta_rules_impl TA) l)"

context ta_inv
begin

lemma ta_rule_compatible_heuristic_subteq: assumes compat: "ta_rule_compatible_heuristic TA u"
    and subt: "l \<unrhd> u"
  shows "ta_rule_compatible TA' (l,r)"
proof -
  {
    fix \<tau>
    assume \<tau>: "\<tau> ` vars_term l \<subseteq> ta_rqs TA'"
    {
      fix x
      assume "x \<in> vars_term u"
      from subt var_imp_supteq[OF this] have "l \<unrhd> Var x" by (rule supteq_trans)
      from var_subt_imp_subt_in_vars[OF this] have "x \<in> vars_term l" .
    }
    with \<tau> have \<tau>u: "\<tau> ` vars_term u \<subseteq> ta_rqs TA'" by auto
    from assms[unfolded ta_rule_compatible_heuristic_def]
    have empty: "rs_\<alpha> (ta_res_list_all (ta_rqs_set TA) (ta_rules_impl TA) u) = {}" by (simp add: rs_correct)
    from ta_res_list_all[OF \<tau>u] empty
    have empty: "ta_res TA' (map_vars \<tau> u) = {}" 
      unfolding ta_res_list by auto
    from subt obtain C where l: "l = C\<langle>u\<rangle>" ..
    have "ta_res TA' (map_vars \<tau> l) = {}"
      unfolding l
    proof (induct C)
      case (More f bef C aft)
      { 
        fix qs
        have "\<And> g. \<exists> i < Suc (length bef + length aft). qs ! i \<notin> (map g bef @ ta_res TA' (map_vars \<tau> C\<langle>u\<rangle>) # map g aft) ! i"
          by (rule exI[of _ "length bef"], simp add: nth_append, insert More, auto)
      }
      thus ?case by (auto simp: o_def)
    qed (insert empty, auto)
  }
  thus ?thesis
    unfolding ta_rule_compatible_def by auto
qed
end


definition ta_trs_compatible_list :: "('q :: linorder,'f :: linorder)ta_impl \<Rightarrow> (('f,'v)rule \<times> 'v list)list \<Rightarrow> (('f,'v)rule \<times> ('f,'q)rule \<times> 'q) check"
where "ta_trs_compatible_list TA R \<equiv> let check = ta_rule_compatible_list TA in
     check_allm (\<lambda> lr. check lr <+? (\<lambda> lq. (fst lr,lq))) R"

context ta_inv
begin
lemma ta_trs_compatible_list: assumes wf_trs: "\<And> l r. (l,r) \<in> fst ` set R \<Longrightarrow> vars_term r \<subseteq> vars_term l"
  and vars: "\<And> l r xs. ((l,r),xs) \<in> set R \<Longrightarrow> xs = vars_term_list l"
  and comp: "isOK(ta_trs_compatible_list TA R)"
  shows "ta_trs_compatible TA' (fst ` set R)" 
proof -
  {
    fix l r xs
    assume rule: "((l,r),xs) \<in> set R"
    from comp[unfolded ta_trs_compatible_list_def Let_def list_all_iff]
      rule have comp: "isOK(ta_rule_compatible_list TA ((l,r),xs))"
      by auto
    from ta_rule_compatible_list[OF wf_trs comp[unfolded vars[OF rule]]] rule
    have "ta_rule_compatible TA' (l,r)" by force
  }
  thus ?thesis unfolding ta_trs_compatible_def 
    by force
qed
end


definition ta_contains_impl :: "('f \<times> nat) list \<Rightarrow> ('q :: linorder,'f :: linorder)ta_impl \<Rightarrow> 'q list \<Rightarrow> ('q + ('f,'q)term)check"
  where "ta_contains_impl F TA qs \<equiv> 
       do { 
         let fin = ta_final_impl TA;
         let look = rm_set_lookup (ta_rules_impl TA);
         let eps = ta_eps_impl TA;
         check_allm (\<lambda> q. check (rs_memb q fin) q) qs
           <+? Inl;
         check_allm (\<lambda> (f,n). let rules = look (f,n) 
              in (check_all (\<lambda> qs'. list_ex (\<lambda> rule. rqs_impl rule = qs' \<and> (let qq = rq_impl rule in list_ex (\<lambda>q. rs_memb q qq) qs)) rules)
          (concat_lists (replicate n qs))) <+? (\<lambda>qs'. Fun f (map Var qs'))) F
            <+? Inr
          } "

context ta_inv
begin
(* I believe that also the other direction is correct, but currently this
   is not needed *)
lemma ta_contains_impl: 
  assumes contain: "isOK(ta_contains_impl F TA qs)"
  shows "ta_contains (set F) TA' (set qs)"
proof -
  note contain = contain[unfolded ta_contains_impl_def Let_def list_ex_iff concat_lists, simplified, simplified rs_correct, simplified]
  have final: "set qs \<subseteq> ta_final TA'" using contain[THEN conjunct1]
    unfolding ta_of_def by (auto simp: rs_correct)
  show ?thesis unfolding ta_contains_def
  proof (rule conjI[OF final], intro allI impI)
    fix f qs'
    assume "(f,length qs') \<in> set F \<and> set qs' \<subseteq> set qs"
    hence f: "(f,length qs') \<in> set F" and i: "\<forall> i < length qs'. qs' ! i \<in> set qs" unfolding set_conv_nth[of qs'] by auto
    from contain[THEN conjunct2, THEN bspec[OF _ f], simplified, THEN spec[of _ qs']] i obtain rule q where q: "q \<in> set qs" and rule: "rule \<in> set (ta_rm_set_lookup (f, length qs'))" 
      and q2: "q \<in> rs_\<alpha> (rq_impl rule)" and qs': "rqs_impl rule = qs'" by auto
    from rm_set_lookup1[OF rule] 
    obtain qsr1 qr qsr2 where rule: "rule = TA_rule_impl f qsr1 qr qsr2"
      and len: "length qs' = length qsr1"
      and eps: "rs_\<alpha> qsr2 = {q'. (qr,q') \<in> (ta_eps TA')^*}"
      and rrule: "f qsr1 \<rightarrow> qr \<in> ta_rules TA'" by auto
    from rule qs' have qs': "qs' = qsr1" by auto
    show "\<exists> q q'. f qs' \<rightarrow> q \<in> ta_rules TA' \<and> q' \<in> set qs \<and> (q,q') \<in> (ta_eps TA')^*"
      by (intro exI conjI, unfold qs', rule rrule, rule q, insert q2[unfolded rule] eps, auto)
  qed
qed
end

datatype ('q,'f)tree_automaton = Tree_Automaton "'q list" "('q,'f)ta_rule list" "('q \<times> 'q)list"


(*    
lemma generate_eps:
  assumes ta_eps: "ta_eps TA = (\<lambda> q. rs_\<alpha> (memo_rbt_rtrancl tran q))" 
  shows "ta_eps_cond TA"
  unfolding ta_eps_cond_def ta_eps memo_rbt_rtrancl by auto
*)
    
fun generate_ta :: "('q :: linorder,'f :: linorder)tree_automaton \<Rightarrow> ('q,'f)ta_impl"
  where "generate_ta (Tree_Automaton fin rules eps) = (let ep = memo_rbt_rtrancl eps;
     rqs_rs = rs_Union (map (\<lambda> rule. (ep (rq rule))) rules);
     rrules = map (conv_ta_rule ep) rules
     in
       \<lparr> ta_final_impl = list_to_rs fin, 
         ta_rules_impl = elem_list_to_rm rsym_impl rrules,
         ta_rqs_impl = rs_to_list rqs_rs,
         ta_rqs_set = rqs_rs,
         ta_eps_impl = ep
          \<rparr>)"

lemma generate_ta_rules: "ta_rules (ta_of (generate_ta (Tree_Automaton fin rules eps))) = set rules"
  (is "ta_rules ?TA' = _")
proof -
  {
    fix eps
    let ?rules = "\<Union> { ta_rule_conv ` {d \<in> set (map (conv_ta_rule eps) rules). rsym_impl d = fn} | fn. True}"
    {
      fix r
      assume "r \<in> ?rules"
      from this[simplified]
      obtain rl rs where r: "r = ta_rule_conv rl"
        and          rl: "rl = conv_ta_rule eps rs"
        and           rs: "rs \<in> set rules"
        by auto
      hence "r \<in> set rules" by (cases rs, auto)
    }
    moreover
    {
      fix r 
      assume "r \<in> set rules"
      then obtain f qs q where r: "r = (f qs \<rightarrow> q)" and rr: "f qs \<rightarrow> q \<in> set rules" by (cases r, auto)
      have "conv_ta_rule eps r = TA_rule_impl f qs q (eps q)" (is "_ = ?rl") unfolding r  by simp
      with rr have rl: "?rl \<in> set (map (conv_ta_rule eps) rules)" (is "_ \<in> ?L") by force
      have r: "r = ta_rule_conv ?rl" unfolding r by auto
      have "r \<in> ta_rule_conv ` ?L" unfolding r using rl by blast
      hence "r \<in> ?rules" by blast
    }
    ultimately have "?rules = set rules" by blast
  }
  thus ?thesis
    unfolding generate_ta.simps  Let_def
    unfolding ta_of_def ta.simps ta_impl.simps elem_list_to_rm.rm_set_lookup
    by auto
qed

lemma generate_ta: "ta_inv(generate_ta TA)"
proof(cases TA)
  case (Tree_Automaton fin rules epsilon)
  obtain eps where eps: "eps = memo_rbt_rtrancl epsilon" by auto
  let ?rqs = "rs_Union (map (\<lambda> rule. (eps (rq rule))) rules)"
  let ?rrules = "map (conv_ta_rule eps) rules"
  from Tree_Automaton eps have A: "generate_ta TA = 
       \<lparr> ta_final_impl = list_to_rs fin, 
         ta_rules_impl = elem_list_to_rm rsym_impl ?rrules,
         ta_rqs_impl = rs_to_list ?rqs,
         ta_rqs_set = ?rqs,
         ta_eps_impl = eps
          \<rparr>" (is "_ = ?TA")
    by (auto simp: Let_def)
  let ?TA' = "ta_of ?TA"
  let ?rules = "\<Union> { ta_rule_conv ` {d \<in> set (map (conv_ta_rule eps) rules). rsym_impl d = fn} | fn. True}"
  have ta_rules: "ta_rules ?TA' = set rules"
    using generate_ta_rules[of fin rules epsilon] 
    by (simp add: eps Let_def)
  have qq: "(\<lambda> (q,q'). (q,q') \<in> (set epsilon)^*) = (set epsilon)^*" unfolding mem_def by auto
  have one: "ta_rqs ?TA' = rs_\<alpha> (ta_rqs_set ?TA)" 
    unfolding ta_rqs_def ta_rules 
    unfolding ta_of_def ta.simps ta_impl.simps rs_Union
    by (simp add: eps memo_rbt_rtrancl qq, auto)
  have two: "set (ta_rqs_impl ?TA) = ta_rqs ?TA'" unfolding one 
    by (auto simp: rs_correct Let_def) 
  show ?thesis unfolding A
    by (unfold_locales, unfold one two, unfold ta_impl.simps elem_list_to_rm.rm_set_lookup ta_rules, 
      auto simp: ta_of_def eps memo_rbt_rtrancl qq)
qed

end
